import numpy as np
import scipy.fft
import scipy.signal
import scipy.ndimage.filters
import scipy.optimize
import scipy.linalg

# Fourier Transform
fft = scipy.fft.fft
fft2 = scipy.fft.fft2
fftn = scipy.fft.fftn
fftshift = scipy.fft.fftshift
ifft = scipy.fft.ifft
ifft2 = scipy.fft.ifft2
ifftn = scipy.fft.ifftn
ifftshift = scipy.fft.ifftshift
dft_matrix = scipy.linalg.dft  # in 'scipy.linalg._special_matrices'
fftfreq = scipy.fft.fftfreq


# todo: Fractional Fourier Transform: pip install git+ssh://git@github.com/audiolabs/python_frft.git#egg=frft


def nextpow2(a):
    """
    Exponent of next higher power of 2. Returns the exponents for the smallest powers
    of two that satisfy 2**p > a

    Parameters
    ----------
    a :     array_like

    Returns
    -------
    p :     array_like

    """

    if np.isscalar(a):
        if a == 0:
            p = 0
        else:
            p = int(np.ceil(np.log2(a)))
    else:
        a = np.asarray(a)
        p = np.zeros(a.shape, dtype=int)
        idx = (a != 0)
        p[idx] = np.ceil(np.log2(a[idx]))

    return p


# Convolution
conv = scipy.signal.convolve
conv2 = scipy.signal.convolve2d
convn = scipy.signal.convolve
deconv = scipy.signal.deconvolve
convolution_matrix = scipy.linalg.convolution_matrix  # in 'scipy.linalg._special_matrices'

# Digital filtering
filter = scipy.signal.lfilter
filtfilt = scipy.signal.filtfilt  # Zero-phase digital filtering
movmedian = scipy.ndimage.filters.median_filter
movmax = scipy.ndimage.filters.maximum_filter
movmin = scipy.ndimage.filters.minimum_filter


def filter2(h, x, shape="full"):
    """
    2-D digital FIR filter

    Parameters
    ----------
    h :     array_like
        The filter, given as a 2D matrix
    x :     array_like
        The data, given as a 2D matrix
    shape : str {'full', 'valid', 'same'}, optional
        A string indicating the size of the output:

        ``full``
           Return the full 2-D filtered data.
        ``valid``
           Return only parts of the filtered data that are computed without zero-padded edges.
        ``same``
           Return the central part of the filtered data, which is the same size as x.

    Returns
    -------
    out :   array_like
        The filtered data
    """

    out = scipy.signal.convolve2d(x, np.rot90(h, 2), mode=shape)
    return out


def movsum(x: np.ndarray, N: int, mode: str = 'same', axis: int = -1) -> np.ndarray:
    """
    Moving sum filter

    Args:
        x:          Input array
        N:          Filter size
        mode:       Mode
        axis:       Axis to filter along

    Returns:
        Filtered array
    """
    kernel = np.ones(N)

    return np.apply_along_axis(lambda m: np.convolve(x, kernel, mode=mode), axis=axis, arr=x)


def movmean(x: np.ndarray, N: int, mode: str = 'same', axis: int = -1) -> np.ndarray:
    """Moving average filter"""

    return movsum(x, N, mode=mode, axis=axis) / N


def movrms(x: np.ndarray, N: int, mode: str = 'same', axis: int = -1) -> np.ndarray:
    """Moving RMS filter"""

    return np.sqrt(movmean(x ** 2, N, mode=mode, axis=axis))


def movvar(x: np.ndarray, N: int, ddof: int = 1, mode: str = 'same', axis: int = -1) -> np.ndarray:
    """Moving variance filter"""

    out = movmean(x ** 2, N, mode=mode, axis=axis) - movmean(x, N, mode=mode, axis=axis) ** 2
    out *= N / (N - ddof)
    return out


def movstd(x: np.ndarray, N: int, mode: str = 'same', axis: int = -1) -> np.ndarray:
    """Moving std filter"""
    out = np.sqrt(movvar(x, N, mode=mode, axis=axis))
    return out


def analyze_window(window, N=64, fs=None, worN=2 ** 17, fftbins=True, plot=False):
    """
    Computes key metrics and optionally plots time- and frequency-domain characteristics of a window function.

    Parameters
    ----------
    window : str | tuple | array_like
        SciPy window spec for `get_window` (e.g., "hann", ("kaiser", 8.6)) or a 1-D array of samples.
    N : int, optional, default=64
        Window length to synthesize if `window` is a spec. Ignored if `window` is array-like.
    fs : float, optional
        Sampling rate [Hz]. If given, values are provided in Hz where applicable.
    worN : int, optional, default=131072
        FFT length for high-resolution spectral measurements (use large, e.g., 2^17).
    fftbins : bool, optional, default=True
        If True, create a "periodic" window, ready to use with `ifftshift` and be multiplied by the result of an FFT
        (see also :func:`~scipy.fft.fftfreq`). If False, create a "symmetric" window, for use in filter design.
    plot : bool, optional, default=False
        If True, produce a magnitude plot with annotations.

    Returns
    -------
    out : dict
        Keys (short, MATLAB-ish naming):
            - 'CG'        : Coherent gain (= mean window amplitude)
            - 'ENBW_bins' : Equivalent Noise Bandwidth in DFT bins (dimensionless)
            - 'ENBW_Hz'   : ENBW in Hz (only if fs is provided)
            - 'MLW'       : Main-lobe width between first nulls (normalized, Nyquist=1)
            - 'MLW_3dB'   : -3 dB bandwidth (normalized, Nyquist=1)
            - 'PSL_dB'    : Peak sidelobe level [dB] (relative to main-lobe peak)
            - 'FSL_dB'    : First sidelobe level [dB] (relative to main-lobe peak)
            - 'ISL_dB'    : Integrated sidelobe level [dB] (power outside main-lobe / inside)
            - 'RO_dBdec'  : Sidelobe roll-off slope [dB/decade] (fit to sidelobe peaks)
            - 'RO_dBoct'  : Same, in dB/octave
            - 'f_axes'    : Dict with normalized frequencies of key points
                           {'f_null': ..., 'f_3dB': ...}  (useful for plotting/verification)
    Notes
    -----
    - ENBW (bins) = mean(w^2) / (mean(w))^2
    - ISL is computed as 10*log10( sum |H|^2 over sidelobes / sum |H|^2 over main lobe ),
      using the single-sided (0..π) spectrum; the ratio is invariant to the bin width.
    - Roll-off is estimated by linear regression of sidelobe *peak* levels vs log10(f).
    """

    # --- get window samples
    if isinstance(window, (str, tuple)):
        w = scipy.signal.get_window(window, N, fftbins=fftbins)
    else:
        w = np.asarray(window, dtype=float)
        w /= w.max()  # normalize to unity gain
        N = w.size

    # --- Coherent Gain (DC gain / mean amplitude)
    CG = w.mean()

    # --- ENBW
    ENBW_bins = np.mean(w ** 2) / (np.mean(w) ** 2)  # In DFT bins
    ENBW_Hz = None
    if fs is not None:
        ENBW_Hz = ENBW_bins * fs / N  # In Hz (if fs is given)

    # --- high-res frequency response (0..pi)
    w_freq, H = scipy.signal.freqz(w, worN=worN, whole=False)  # rad/sample, 0..pi
    freqs = w_freq / np.pi  # normalized (Nyquist = 1)
    if fs is not None:
        freqs *= fs / 2  # in Hz if fs is given
    H_mag = np.abs(H)
    H_mag /= H_mag.max()  # normalize to 1 at peak
    H_dB = 20 * np.log10(np.maximum(H_mag, 1e-300))  # dB-safe

    # --- find first *minimum* after DC (first null)
    # use minima of magnitude: peaks of -H_mag
    minima_idx, _ = scipy.signal.find_peaks(-H_mag)
    if minima_idx.size == 0:
        # fallback: first index where it stops decreasing
        deriv = np.diff(H_mag)
        k_null = np.argmax(deriv > 0) or 1
    else:
        k_null = minima_idx[0]
    f_null = freqs[k_null]
    MLW = 2 * f_null  # width between ±f_null in normalized frequency

    # --- -3 dB bandwidth (interpolated)
    # find first crossing to -3 dB before first null
    mask_main = np.arange(k_null + 1)
    k_hp = np.where(H_dB[mask_main] <= -3.0)[0]
    if k_hp.size:
        k2 = k_hp[0]
        k1 = max(k2 - 1, 0)
        # linear interpolation on dB scale
        y1, y2 = H_dB[k1], H_dB[k2]
        x1, x2 = freqs[k1], freqs[k2]
        if y2 == y1:
            f_3dB = x2
        else:
            f_3dB = x1 + (x2 - x1) * ((-3.0 - y1) / (y2 - y1))
        MLW_3dB = 2 * f_3dB
    else:
        f_3dB, MLW_3dB = np.nan, np.nan

    # --- sidelobe peaks and levels
    peaks_idx, _ = scipy.signal.find_peaks(H_mag)
    side_peaks = peaks_idx[freqs[peaks_idx] > f_null]
    if side_peaks.size:
        PSL_dB = H_dB[side_peaks].max()  # max sidelobe
        FSL_dB = H_dB[side_peaks[0]]  # first sidelobe
    else:
        PSL_dB, FSL_dB = np.nan, np.nan

    # --- ISL (power ratio sidelobes/main-lobe)
    P = H_mag ** 2
    P_main = P[:k_null + 1].sum()
    P_side = P[k_null + 1:].sum()
    ISL_dB = 10 * np.log10(P_side / P_main) if P_main > 0 else np.nan

    # --- roll-off: slope of sidelobe *peaks* in dB vs log10(freq)
    if side_peaks.size >= 2:
        fpk = freqs[side_peaks]
        ypk = H_dB[side_peaks]
        # ignore any numerically tiny/late peaks
        good = (fpk > f_null) & np.isfinite(ypk) & (ypk < -1)  # below -1 dB to avoid main-lobe skirt
        fpk, ypk = fpk[good], ypk[good]
        if fpk.size >= 2:
            x = np.log10(fpk)
            a, b = np.polyfit(x, ypk, 1)  # y ≈ a*log10(f) + b
            RO_dBdec = float(a)
            RO_dBoct = float(a * np.log10(2.0))  # per octave
        else:
            RO_dBdec = RO_dBoct = np.nan
    else:
        RO_dBdec = RO_dBoct = np.nan

    out = dict(
            CG=CG,
            ENBW_bins=ENBW_bins,
            ENBW_Hz=ENBW_Hz,
            MLW=MLW,
            MLW_3dB=MLW_3dB,
            PSL_dB=PSL_dB,
            FSL_dB=FSL_dB,
            ISL_dB=ISL_dB,
            RO_dBdec=RO_dBdec,
            RO_dBoct=RO_dBoct,
            f_null=f_null,
            f_3dB=f_3dB,
    )

    if plot:
        import matplotlib.pyplot as plt
        fig, axs = plt.subplots(1, 2, figsize=(12, 6))
        axs = iter(axs)

        # ---- Time-domain
        ax = next(axs)
        ax.plot(np.arange(N), w)

        ax.set_xlabel("Samples")
        ax.set_ylabel("Amplitude")

        # Build window label for title (if name/spec provided)
        win_label = f"{N=}"
        if isinstance(window, str):
            win_label = f"{window}({N=})"
        elif isinstance(window, tuple) and len(window) >= 1:
            name = str(window[0])
            params = ", ".join(f"{p:g}" if isinstance(p, (int, float)) else str(p) for p in window[1:])
            win_label = f"{name}({N=}, {params})" if params else name
        ax.set_title(f"Time Domain — {win_label if win_label else ''}")

        ax.set_ylim(0, w.max() * 1.05)

        # ---- Frequency domain
        ax = next(axs)
        ax.plot(freqs, H_dB)

        xlabel = "Normalized frequency (× Nyquist)"
        if fs is not None:
            xlabel = "Frequency (Hz)"
        ax.set_xlabel(xlabel)
        ax.set_ylabel("Magnitude (dB)")

        ax.set_title(f"Frequency Domain")
        ax.grid(True)

        # annotate main-lobe width
        ax.axvline(+f_null, linestyle="--", linewidth=0.5)
        ax.axvline(-f_null, linestyle="--", linewidth=0.5)
        ax.text(f_null, -6, rf"$\text{{MLW}}={MLW:.3f}${'' if fs is None else 'Hz'}", ha="left", va="top")

        # annotate 3dB bandwidth
        if np.isfinite(f_3dB):
            ax.axvline(f_3dB, linestyle=":")
            ax.text(f_3dB, -3, rf"$\text{{MLW}}_{{-3\text{{dB}}}}={MLW_3dB:.3f}${'' if fs is None else 'Hz'}",
                    ha="left", va="bottom")

        # mark PSL peak
        if side_peaks.size:
            k_psl = side_peaks[np.argmax(H_dB[side_peaks])]
            ax.plot(freqs[k_psl], H_dB[k_psl], marker="o")
            ax.text(freqs[k_psl], H_dB[k_psl], rf"PSL$={H_dB[k_psl]:.1f}$dB", ha="left", va="bottom")

        # shade sidelobe region for ISL visualization
        ax.fill_between(freqs[k_null + 1:], H_dB[k_null + 1:], -300, alpha=0.08)

        # Metrics textbox
        unit_freq = "" if fs is None else " Hz"
        metrics_text = "\n".join([f"{k}={v:.3f}" for k, v in out.items() if v is not None])
        ax.text(0.98, 0.98,
                metrics_text,
                fontsize="x-small",
                ha="right", va="top",
                transform=ax.transAxes,
                bbox=dict(boxstyle="round", facecolor="white", alpha=0.8, edgecolor="0.8"),
        )

        ax.set_xlim(0, 1 if fs is None else fs / 2)
        ax.set_ylim(-150, 5)

        plt.tight_layout()
        plt.show(block=False)

    return out


def analyze_filter(b=None, a=1, sos=None, fs=None, worN=2 ** 14,
        passband_tol=3.0, stopband_tol=60.0, eps=1e-20, plot=False):
    """
    Analyze key characteristics of a digital filter.

    Parameters
    ----------
    b, a : array_like, optional
        Numerator (b) and denominator (a) coefficients of the filter.
        Ignored if sos is provided.
    sos : array_like, optional
        Second-order sections representation of the IIR filter.
        Should be an array of shape (n_sections, 6) where each row
        represents one second-order section in the format [b0, b1, b2, a0, a1, a2].
        If provided, b and a parameters are ignored.
    fs : float, optional
        Sampling rate [Hz]. If None, frequencies are normalized to Nyquist (0..1).
    worN : int, optional
        Number of frequency samples for freqz. Default 2^14.
    passband_tol : float
        Passband edge defined at -passband_tol [dB] (default 3 dB).
    stopband_tol : float
        Stopband defined at -stopband_tol [dB] (default 60 dB).
    plot : bool, optional
        If True, plot responses.

    Returns
    -------
    out : dict
        Keys:
            - 'order'        : Filter order
            - 'type'         : FIR/IIR guess
            - 'bw_3dB'       : -3 dB bandwidth
            - 'f_stop'       : Stopband edge frequency
            - 'tw'           : Transition width = f_stop - f_pass
            - 'atten_stop'   : Min stopband attenuation [dB]
            - 'ripple_pass'  : Passband ripple [dB]
            - 'group_delay_mean': Mean group delay [samples]
            - 'group_delay_var' : Variance of group delay
            - 'stability'    : Stable (all poles inside unit circle)
            - 'f'            : Frequency axis
            - 'H'            : Frequency response (complex)
            - 'n_sections'   : Number of sections (only for SOS)
    """

    # Determine filter representation and convert if necessary
    if not ((b is None) ^ (sos is None)):
        raise ValueError("Either `b` or `sos` must be provided.")

    is_sos = b is None

    if is_sos:  # sos
        sos = np.asarray(sos)
        if sos.ndim != 2 or sos.shape[1] != 6:
            raise ValueError("`sos` must have shape (n_sections, 6)")

        n_sections = sos.shape[0]

        # Convert SOS to transfer function for some calculations
        b, a = scipy.signal.sos2tf(sos)
        b, a = np.atleast_1d(b), np.atleast_1d(a)

        # Order calculation for SOS
        order = n_sections * 2  # Each section is 2nd order
        ftype = "IIR"  # SOS is always IIR

    else:  # b, a
        b, a = np.atleast_1d(b), np.atleast_1d(a)
        n_sections = None

        # Order and type
        order = max(len(b), len(a)) - 1
        ftype = "FIR" if np.allclose(a, [1]) else "IIR"

    # Frequency response - use appropriate function for SOS vs b,a
    if fs is None:  # Normalized frequency [0..1], with Nyquist=1
        if is_sos:
            w, H = scipy.signal.sosfreqz(sos, worN=worN, whole=False)
        else:
            w, H = scipy.signal.freqz(b, a, worN=worN, whole=False)
        f = w / np.pi
    else:
        if is_sos:
            f, H = scipy.signal.sosfreqz(sos, worN=worN, fs=fs)
        else:
            f, H = scipy.signal.freqz(b, a, worN=worN, fs=fs)

    H_mag = np.abs(H)
    H_dB = 20 * np.log10(H_mag + eps)

    # --- Passband edge (first -3 dB crossing)
    idx_pass = np.argmax(H_dB <= -passband_tol)
    f_pass = f[idx_pass] if idx_pass > 0 else np.nan
    bw_3dB = f_pass

    # Passband ripple (up to f_pass)
    ripple_pass = (H_dB[:idx_pass].max() - H_dB[:idx_pass].min()) if idx_pass > 0 else np.nan

    # --- Stopband edge (first frequency below stopband_tol)
    idx_stop = np.argmax(H_dB <= -stopband_tol)
    f_stop = f[idx_stop] if idx_stop > 0 else np.nan

    # Stopband attenuation (after f_stop)
    atten_stop = -H_dB[idx_stop:].max() if idx_stop > 0 else np.nan

    # Transition width
    tw = (f_stop - f_pass) if np.isfinite(f_pass) and np.isfinite(f_stop) else np.nan

    # --- Group delay
    try:
        system = sos if is_sos else (b, a)
        w_gd, gd = scipy.signal.group_delay(system, worN=worN, fs=(fs if fs else 2 * np.pi))
        group_delay_mean = np.mean(gd[np.isfinite(gd)])
        group_delay_var = np.var(gd[np.isfinite(gd)])
    except Exception:
        group_delay_mean = group_delay_var = np.nan

    # --- Stability
    if is_sos:
        # Check stability for each section in SOS
        stability = True
        all_zeros, all_poles = [], []
        for section in sos:
            b_sec = section[:3]  # [b0, b1, b2]
            a_sec = section[3:]  # [a0, a1, a2]
            # Normalize by a0
            if a_sec[0] != 0:
                a_sec = a_sec / a_sec[0]
                b_sec = b_sec / section[3]

            zeros_sec = np.roots(b_sec)
            poles_sec = np.roots(a_sec)

            all_zeros.extend(zeros_sec)
            all_poles.extend(poles_sec)

            # Check if any pole is outside unit circle
            if np.any(np.abs(poles_sec) >= 1):
                stability = False

        zeros = np.array(all_zeros)
        poles = np.array(all_poles)
    else:
        zeros, poles = np.roots(b), np.roots(a)
        stability = np.all(np.abs(poles) < 1)

    out = dict(
            order=order,
            type=ftype,
            bw_3dB=bw_3dB,
            f_stop=f_stop,
            tw=tw,
            atten_stop=atten_stop,
            ripple_pass=ripple_pass,
            group_delay_mean=group_delay_mean,
            group_delay_var=group_delay_var,
            stability=stability,
            f=f,
            H=H
    )
    if is_sos:
        out['n_sections'] = n_sections

    if plot:
        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(2, 2, figsize=(12, 8))
        axs = iter(axs.flatten())

        # Impulse response
        if ftype == "FIR":
            h = b
        else:  # IIR
            n_imp = max(200, order * 4)
            x = np.zeros(n_imp)
            x[0] = 1
            if is_sos:
                h = scipy.signal.sosfilt(sos, x)
            else:
                h = scipy.signal.lfilter(b, a, x)
        ax = next(axs)
        ax.stem(np.arange(len(h)), h)
        title = f"Impulse Response"
        if is_sos:
            title += f" (SOS: {n_sections} sections)"
        ax.set_title(title)
        ax.set_xlabel("Samples")
        ax.set_ylabel("Amplitude")
        ax.grid(True)

        # Pole-zero plot
        ax = next(axs)
        ax.scatter(np.real(zeros), np.imag(zeros), marker='o', label='Zeros')
        ax.scatter(np.real(poles), np.imag(poles), marker='x', label='Poles')
        circ = plt.Circle((0, 0), 1, color='black', fill=False, ls='--')
        ax.add_patch(circ)
        ax.set_title("Pole-Zero Plot")
        ax.set_xlabel(r"$\Re{\{z\}}$")
        ax.set_ylabel(r"$\Im{\{z\}}$")
        ax.set_aspect('equal')
        ax.legend()
        ax.grid(True)

        # Magnitude response
        ax = next(axs)
        ax.plot(f, H_dB)
        ax.set_title("Magnitude Response")
        if fs is None:
            ax.set_xlabel("Normalized Frequency (× Nyquist)")
            ax.set_xlim(0, 1)
        else:
            ax.set_xlabel("Frequency [Hz]")
            ax.set_xlim(0, fs / 2)
        ax.set_ylabel("Magnitude [dB]")
        ax.grid(True)
        if np.isfinite(f_pass):
            ax.axvline(f_pass, ls="--", c="r", label=f"Passband edge {f_pass:.3g}")
        if np.isfinite(f_stop):
            ax.axvline(f_stop, ls="--", c="g", label=f"Stopband edge {f_stop:.3g}")
        ax.legend()

        # Phase response
        ax = next(axs)
        ax.plot(f, np.rad2deg(np.unwrap(np.angle(H))))
        ax.set_title("Phase Response")
        ax.set_ylabel("Phase [deg]")
        if fs is None:
            ax.set_xlabel("Normalized Frequency (× Nyquist)")
            ax.set_xlim(0, 1)
        else:
            ax.set_xlabel("Frequency [Hz]")
            ax.set_xlim(0, fs / 2)
        ax.grid(True)

        plt.tight_layout()
        plt.show(block=False)

    return out



def to_uint8(*args):
    """Convert a float value in the range [0, 1] to an unsigned 8-bit integer."""
    out = [0] * len(args)
    for i, arg in enumerate(args):
        if int(arg) == arg and 0 <= arg <= 255:
            out[i] = int(arg)
        elif 0 <= arg <= 1:
            out[i] = int(arg * 255)
        else:
            raise ValueError(f"Value {arg} is out of range [0, 1] or [0, 255] for conversion to uint8.")
    return tuple(out)


def print_in_color(text, foreground_rgb=None, background_rgb=None):
    out = text
    if foreground_rgb is not None:
        foreground_rgb = to_uint8(*foreground_rgb)
        foreground_rgb = [str(c) for c in foreground_rgb]
        out = f"\x1b[38;2;{';'.join(foreground_rgb)}m{text}"
    if background_rgb is not None:
        background_rgb = to_uint8(*background_rgb)
        background_rgb = [str(c) for c in background_rgb]
        out = f"\x1b[48;2;{';'.join(background_rgb)}m{text}"

    return f"{out}\x1b[0m"  # Reset all styles



import pandas as pd


def load_json(file):
	# TODO replace with pd.read_json
	import json

	with open(file) as f:
		d = json.load(f)
	return d


def write_json(d, file, *args, **kwargs):
	kwargs = {"indent": 4} | kwargs
	# dump = json.dumps(d, *args, **kwargs)
	# with open(file, 'w') as f:
	# 	json.dump(dump, f)
	pd.Series(d).to_json(file, *args, **kwargs)



import os
import sys


def import_module(file_name):
	"""
	Import a module/file into Python session.


	Args:
		file_name:   file name

	Returns:

	"""
	dir, file_name = os.path.split(file_name)
	sys.path.append(dir)
	h = __import__(file_name)
	sys.path.pop(-1)
	return h



__all__ = ['BaseProgressBar', 'TextProgressBar',
	'EnhancedTextProgressBar', 'TqdmProgressBar',
	'HTMLProgressBar', 'tqdm_']

import time
import datetime
import sys
from collections.abc import Iterable, Callable
from tqdm import tqdm


class BaseProgressBar(object):
	"""
	An abstract progress bar with some shared functionality.

	Example usage:

		n_vec = linspace(0, 10, 100)
		pbar = TextProgressBar(len(n_vec))
		for n in n_vec:
			pbar.update()
			compute_with_n(n)
		pbar.finished()

	"""

	def __init__(self, iterations=0, chunk_size=10, **kwargs):
		self._start(iterations, chunk_size)

	def _start(self, iterations, chunk_size=10, **kwargs):
		self.N = float(iterations)
		self.n = 0
		self.p_chunk_size = chunk_size
		self.p_chunk = chunk_size
		self.t_start = time.time()
		self.t_done = self.t_start - 1

	def update(self):
		pass

	def total_time(self):
		return self.t_done - self.t_start

	def time_elapsed(self):
		return "%6.2fs" % (time.time() - self.t_start)

	def time_remaining_est(self, p):
		if 100 >= p > 0.0:
			t_r_est = (time.time() - self.t_start) * (100.0 - p) / p
		else:
			t_r_est = 0

		dd = datetime.datetime(1, 1, 1) + datetime.timedelta(seconds=t_r_est)
		time_string = "%02d:%02d:%02d:%02d" % \
		              (dd.day - 1, dd.hour, dd.minute, dd.second)

		return time_string

	def finished(self):
		self.t_done = time.time()


class TextProgressBar(BaseProgressBar):
	"""
	A simple text-based progress bar.
	"""

	def __init__(self, iterations=0, chunk_size=10, **kwargs):
		super()._start(iterations, chunk_size)

	def update(self):
		self.n += 1
		n = self.n
		p = (n / self.N) * 100.0
		if p >= self.p_chunk:
			print("%4.1f%%." % p +
			      " Run time: %s." % self.time_elapsed() +
			      " Est. time left: %s" % self.time_remaining_est(p))
			sys.stdout.flush()
			self.p_chunk += self.p_chunk_size

	def finished(self):
		self.t_done = time.time()
		print("Total run time: %s" % self.time_elapsed())


class EnhancedTextProgressBar(BaseProgressBar):
	"""
	An enhanced text-based progress bar.
	"""

	def __init__(self, iterations=0, chunk_size=10, **kwargs):
		super()._start(iterations, chunk_size)
		self.fill_char = '*'
		self.width = 25

	def update(self):
		self.n += 1
		n = self.n
		percent_done = int(round(n / self.N * 100.0))
		all_full = self.width - 2
		num_hashes = int(round((percent_done / 100.0) * all_full))
		prog_bar = ('[' + self.fill_char * num_hashes +
		            ' ' * (all_full - num_hashes) + ']')
		pct_place = (len(prog_bar) // 2) - len(str(percent_done))
		pct_string = '%d%%' % percent_done
		prog_bar = (prog_bar[0:pct_place] +
		            (pct_string + prog_bar[pct_place + len(pct_string):]))
		prog_bar += ' Elapsed %s / Remaining %s' % (
			self.time_elapsed().strip(),
			self.time_remaining_est(percent_done))
		print('\r', prog_bar, end='')
		sys.stdout.flush()

	def finished(self):
		self.t_done = time.time()
		print("\r", "Total run time: %s" % self.time_elapsed())


class TqdmProgressBar(BaseProgressBar):
	"""
	A progress bar using tqdm module
	"""

	def __init__(self, iterations=0, chunk_size=10, **kwargs):
		from tqdm.auto import tqdm
		self.pbar = tqdm(total=iterations, **kwargs)
		self.t_start = time.time()
		self.t_done = self.t_start - 1

	def update(self):
		self.pbar.update()

	def finished(self):
		self.pbar.close()
		self.t_done = time.time()


class HTMLProgressBar(BaseProgressBar):
	"""
	A simple HTML progress bar for using in IPython notebooks. Based on
	IPython ProgressBar demo notebook:
	https://github.com/ipython/ipython/tree/master/examples/notebooks

	Example usage:

		n_vec = linspace(0, 10, 100)
		pbar = HTMLProgressBar(len(n_vec))
		for n in n_vec:
			pbar.update()
			compute_with_n(n)
	"""

	def __init__(self, iterations=0, chunk_size=1.0, **kwargs):
		from IPython.display import HTML, Javascript, display
		import uuid

		self.display = display
		self.Javascript = Javascript
		self.divid = str(uuid.uuid4())
		self.textid = str(uuid.uuid4())
		self.pb = HTML(
				'<div style="border: 2px solid grey; width: 600px">\n  '
				f'<div id="{self.divid}" '
				'style="background-color: rgba(121,195,106,0.75); '
				'width:0%">&nbsp;</div>\n'
				'</div>\n'
				f'<p id="{self.textid}"></p>\n'
		)
		self.display(self.pb)
		super()._start(iterations, chunk_size)

	def update(self):
		self.n += 1
		n = self.n
		p = (n / self.N) * 100.0
		if p >= self.p_chunk:
			lbl = ("Elapsed time: %s. " % self.time_elapsed() +
			       "Est. remaining time: %s." % self.time_remaining_est(p))
			js_code = ("$('div#%s').width('%i%%');" % (self.divid, p) +
			           "$('p#%s').text('%s');" % (self.textid, lbl))
			self.display(self.Javascript(js_code))
			self.p_chunk += self.p_chunk_size

	def finished(self):
		self.t_done = time.time()
		lbl = "Elapsed time: %s" % self.time_elapsed()
		js_code = ("$('div#%s').width('%i%%');" % (self.divid, 100.0) +
		           "$('p#%s').text('%s');" % (self.textid, lbl))
		self.display(self.Javascript(js_code))


def tqdm_(iterable: Iterable,
		desc: str | Callable = None,
		total: int | float = None,
		disable: bool = False,
		unit: str = 'it',
		postfix: dict | Callable = None,
		**kwargs
):
	"""
	    Decorate an iterable object, returning an iterator which acts exactly
	    like the original iterable, but prints a dynamically updating
	    progressbar every time a value is requested.

	    Parameters
	    ----------
	    iterable  : iterable, optional
	        Iterable to decorate with a progressbar.
	        Leave blank to manually manage the updates.
	    desc  : str, optional
	        Prefix for the progressbar.
	    total  : int or float, optional
	        The number of expected iterations. If unspecified,
	        len(iterable) is used if possible. If float("inf") or as a last
	        resort, only basic progress statistics are displayed
	        (no ETA, no progressbar).
	        If `gui` is True and this parameter needs subsequent updating,
	        specify an initial arbitrary large positive number,
	        e.g. 9e9.
	    leave  : bool, optional
	        If [default: True], keeps all traces of the progressbar
	        upon termination of iteration.
	        If `None`, will leave only if `position` is `0`.
	    file  : `io.TextIOWrapper` or `io.StringIO`, optional
	        Specifies where to output the progress messages
	        (default: sys.stderr). Uses `file.write(str)` and `file.flush()`
	        methods.  For encoding, see `write_bytes`.
	    ncols  : int, optional
	        The width of the entire output message. If specified,
	        dynamically resizes the progressbar to stay within this bound.
	        If unspecified, attempts to use environment width. The
	        fallback is a meter width of 10 and no limit for the counter and
	        statistics. If 0, will not print any meter (only stats).
	    mininterval  : float, optional
	        Minimum progress display update interval [default: 0.1] seconds.
	    maxinterval  : float, optional
	        Maximum progress display update interval [default: 10] seconds.
	        Automatically adjusts `miniters` to correspond to `mininterval`
	        after long display update lag. Only works if `dynamic_miniters`
	        or monitor thread is enabled.
	    miniters  : int or float, optional
	        Minimum progress display update interval, in iterations.
	        If 0 and `dynamic_miniters`, will automatically adjust to equal
	        `mininterval` (more CPU efficient, good for tight loops).
	        If > 0, will skip display of specified number of iterations.
	        Tweak this and `mininterval` to get very efficient loops.
	        If your progress is erratic with both fast and slow iterations
	        (network, skipping items, etc) you should set miniters=1.
	    ascii  : bool or str, optional
	        If unspecified or False, use unicode (smooth blocks) to fill
	        the meter. The fallback is to use ASCII characters " 123456789#".
	    disable  : bool, optional
	        Whether to disable the entire progressbar wrapper
	        [default: False]. If set to None, disable on non-TTY.
	    unit  : str, optional
	        String that will be used to define the unit of each iteration
	        [default: it].
	    unit_scale  : bool or int or float, optional
	        If 1 or True, the number of iterations will be reduced/scaled
	        automatically and a metric prefix following the
	        International System of Units standard will be added
	        (kilo, mega, etc.) [default: False]. If any other non-zero
	        number, will scale `total` and `n`.
	    dynamic_ncols  : bool, optional
	        If set, constantly alters `ncols` and `nrows` to the
	        environment (allowing for window resizes) [default: False].
	    smoothing  : float, optional
	        Exponential moving average smoothing factor for speed estimates
	        (ignored in GUI mode). Ranges from 0 (average speed) to 1
	        (current/instantaneous speed) [default: 0.3].
	    bar_format  : str, optional
	        Specify a custom bar string formatting. May impact performance.
	        [default: '{l_bar}{bar}{r_bar}'], where
	        l_bar='{desc}: {percentage:3.0f}%|' and
	        r_bar='| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, '
	            '{rate_fmt}{postfix}]'
	        Possible vars: l_bar, bar, r_bar, n, n_fmt, total, total_fmt,
	            percentage, elapsed, elapsed_s, ncols, nrows, desc, unit,
	            rate, rate_fmt, rate_noinv, rate_noinv_fmt,
	            rate_inv, rate_inv_fmt, postfix, unit_divisor,
	            remaining, remaining_s, eta.
	        Note that a trailing ": " is automatically removed after {desc}
	        if the latter is empty.
	    initial  : int or float, optional
	        The initial counter value. Useful when restarting a progress
	        bar [default: 0]. If using float, consider specifying `{n:.3f}`
	        or similar in `bar_format`, or specifying `unit_scale`.
	    position  : int, optional
	        Specify the line offset to print this bar (starting from 0)
	        Automatic if unspecified.
	        Useful to manage multiple bars at once (eg, from threads).
	    postfix  : dict or *, optional
	        Specify additional stats to display at the end of the bar.
	        Calls `set_postfix(**postfix)` if possible (dict).
	    unit_divisor  : float, optional
	        [default: 1000], ignored unless `unit_scale` is True.
	    write_bytes  : bool, optional
	        Whether to write bytes. If (default: False) will write unicode.
	    lock_args  : tuple, optional
	        Passed to `refresh` for intermediate output
	        (initialisation, iterating, and updating).
	    nrows  : int, optional
	        The screen height. If specified, hides nested bars outside this
	        bound. If unspecified, attempts to use environment height.
	        The fallback is 20.
	    colour  : str, optional
	        Bar colour (e.g. 'green', '#00ff00').
	    delay  : float, optional
	        Don't display until [default: 0] seconds have elapsed.
	    gui  : bool, optional
	        WARNING: internal parameter - do not use.
	        Use tqdm.gui.tqdm(...) instead. If set, will attempt to use
	        matplotlib animations for a graphical output [default: False].

	    Returns
	    -------
	    out  : decorated iterator.
	    """

	if not callable(desc):
		desc_str = desc
		desc = lambda i: desc_str
	if not callable(postfix):
		postfix = lambda i: None

	if hasattr(iterable, '__len__'):
		total = len(iterable)

	pbar = tqdm(total=total, disable=disable, unit=unit, **kwargs)

	for i, iter in enumerate(iterable):
		pbar.set_description(desc(i))
		pbar.set_postfix(postfix(i))
		yield iter
		pbar.update(1)



import numpy as np
import scipy.stats
from collections import namedtuple

from ..uncertainties_math import val, to_numpy

Power_divergenceResult = namedtuple('Power_divergenceResult',
		('statistic', 'pvalue'))


def chi_squared_test(
		f_exp, f_obs,
		f_obs_err=None,
		ddof: int = 0,
		reduced: bool = False):
	"""
	Compute the chi-squared statistic and p-value for the goodness-of-fit test (see ref. [1]).
	Same as scipy.stats.chisquare(f_obs, f_exp, ddof)

    Parameters
    ----------
    f_exp :     array-like
        Expected (theoretical) values.
    f_obs :     array-like
        Observed values.
    f_obs_err : array-like, optional
    	Uncertainties in the observed values.
    ddof :      int, optional
        Degrees of freedom adjustment (e.g., number of fitted parameters). Default is 0.
    reduced :   bool, optional
        If True, the reduced chi-squared test is performed (see ref. [2]). Default is False.
        - chi2 >> 1 indicates a poor model fit.
        - chi2 > 1 indicates that the fit has not fully captured the data (or that the error variance has been
                underestimated).
        - A value of chi2 ~ 1 around indicates that the extent of the match between observations and estimates is
                in accord with the error variance.
        - chi2 < 1 indicates that the model is overfitting the data (either the model is improperly fitting noise,
                or the error variance has been overestimated).

    Returns
    -------
    chi2 : float
        Chi-squared statistic.
    p_value : float
        p-value for the goodness-of-fit test.

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Pearson%27s_chi-squared_test
    [2] https://en.wikipedia.org/wiki/Reduced_chi-squared_statistic
	"""

	f_exp, f_exp_err = to_numpy(f_exp)
	f_obs, f_obs_err = to_numpy(f_obs, f_obs_err)

	if reduced:
		assert f_obs_err is not None, "f_exp_err, f_obs_err must be provided."
		if f_exp_err is None:
			f_exp_err = 0
		sigma2 = np.sqrt(f_exp_err ** 2 + f_obs_err ** 2)
		chi2 = np.sum((f_obs - f_exp) ** 2 / sigma2 ** 2)
	else:
		f_exp /= f_exp.sum()
		f_obs /= f_obs.sum()
		chi2 = np.sum((f_obs - f_exp) ** 2 / f_exp)

	dof = len(f_exp) - 1 - ddof
	p_value = scipy.stats.chi2.sf(val(chi2), dof)

	return Power_divergenceResult(chi2, p_value)



import sympy


class SymExpr:

	def __init__(self, expression, subs_value=None, dev=None, units=None, info=None):
		"""

		Args:
			expression:
			subs_value:     dict {'x': 5, 'y': 6}
			dev:            dict{'x': .5, 'y': .6}
			units:          str
			info:           str
		"""

		self.expr = expression

		self.subs_value = subs_value
		self.dev = dev
		self.units = units
		self.info = info

	def __call__(self):
		return self.expr

	def subs(self, out_type=sympy.Expr):
		"""

		Args:
			out_type: sympy.Expr or str

		Returns:

		"""
		assert out_type in (sympy.Expr, str)
		value = self.expr
		dev = self.dev

		if isinstance(self.expr, sympy.Expr):
			if self.subs_value is None:
				print("No values to substitute.")
			else:
				value = self.expr.subs(self.subs_value)
				dev = self.calc_dev()

		if out_type is sympy.Expr:
			out = (value, dev)
		else:
			out = value.__str__()

			if dev is not None:
				out += f' +- {dev.__str__()}'

			if self.units is not None:
				out += f' [{self.units}]'

		return out

	def calc_dev(self, subs_value=True):
		gradient = dict()
		for sym in self.expr.free_symbols:
			gradient[sym] = self.expr.diff(sym)

		out = sympy.core.numbers.Zero()
		for sym in self.expr.free_symbols:
			if sym in self.dev.keys() and self.dev[sym] is not None:
				out += (gradient[sym] * self.dev[sym]) ** 2

		out = sympy.sqrt(out)

		if subs_value and self.subs_value is not None:
			out = out.subs(self.subs_value)

		return out

	def __str__(self):
		return f'{self.expr}, dev={self.dev}'

	def __repr__(self):
		return self.__str__()



from datetime import datetime

time_str_formats = {
	"full":              "%Y-%m-%d %H:%M:%S,%f",  # 2023-07-11 00:33:28,881542
	"full_dash":  "%Y-%m-%d %H-%M-%S-%f",  # 2023-07-11 00-33-28
	"yyyy-mm-dd": "%Y-%m-%d",  # 2023-07-11
	"dd/mm/yyyy":        "%d/%m/%Y",  # 11/07/2023
	"mm/dd/yyyy":        "%m/%d/%Y",  # 07/11/2023
	"d, m_name, y":      "%b %d, %Y",  # Jul 11, 2023
	"d_name, m_name, y": "%a, %B %d, %Y",  # Tue, July 11, 2023
}


def get_time():
	"""
	Get time a datetime object

	Returns:

	"""
	return datetime.now()


def get_time_str(fmt=time_str_formats["full_dash"]) -> str:
	"""
	Get time as a string with a specified format.

	Args:
		fmt:        Format of the time string

	Returns:

	"""
	return datetime.now().strftime(fmt)



"""Liron Utils - A collection of utility modules."""

__version__ = "2025.09.11"

# Import all submodules
from . import graphics
from . import signal_processing
from . import machine_learning
from . import pure_python
from . import symbolic_math
from . import time
from . import uncertainties_math
from . import files
from . import web
from . import manim_animations

__all__ = [
    '__version__',
    'graphics',
    'signal_processing',
    'machine_learning',
    'pure_python',
    'symbolic_math',
    'time',
    'uncertainties_math',
    'files',
    'web',
    'manim_animations'
]



import numpy as np
import scipy.stats
import scipy.optimize
import scipy.odr as odr
import scipy.signal
from collections import namedtuple

from .stats import chi_squared_test
from ..uncertainties_math import val, to_numpy, from_numpy
from .. import graphics as gr

linear_regression = scipy.stats.linregress


def linear_fit(x, y, xerr=None, yerr=None, beta0=None, **kwargs):
	"""
	Linear regression

	Parameters
	----------
	x, y :              array_like
		Data
	xerr, yerr :        array_like, optional
		Deviations in 'x','y'. 'x','y' may also be sent as 'uncertainties' arrays, then 'xerr','yerr' are disregarded
	beta0 :             array_like, optional
		Initial parameter values, sent as a 2-tuple '(a, b)', where 'a' is the slope and 'b' is the intercept
	kwargs :            sent to odr.ODR

	Returns
	-------

	"""
	x, xerr = to_numpy(x, xerr)
	y, yerr = to_numpy(y, yerr)

	if xerr is not None:
		xerr[xerr == 0] = np.nan
	if yerr is not None:
		yerr[yerr == 0] = np.nan

	if beta0 is None:
		beta0 = [0, 0]  # [slope, intercept]

	def f(B, x):
		return B[0] * x + B[1]

	linear_model = odr.Model(f)
	data = odr.RealData(x, y, xerr, yerr)

	out = odr.ODR(data, linear_model, beta0, **kwargs).run()

	# Calculate params
	from uncertainties import ufloat
	slope = ufloat(out.beta[0], out.sd_beta[0])
	intercept = ufloat(out.beta[1], out.sd_beta[1])

	pred = lambda x: f([slope, intercept], x)

	y_pred = pred(x)

	from sklearn.metrics import r2_score

	r2 = r2_score(y, val(y_pred))

	Linear_regression = namedtuple("Linear_regression",
			["h", "eval", "slope", "intercept", "r2"])
	return Linear_regression(h=out, eval=pred, slope=slope, intercept=intercept, r2=r2)


def curve_fit(fit_fcn,
		x, y, p0=None,
		xerr=None, yerr=None,
		plot=False, plot_data_and_curve_fit_kw=None,
		**kwargs):
	"""
	Use non-linear least squares to fit a function, f, to data.
	Assumes ``y = f(x, *params) + eps``.

	Parameters
	----------
	fit_fcn :   callable
				The model function, f(x, ...). It must take the independent
				variable as the first argument and the parameters to fit as
				separate remaining arguments
	x :         array_like
				The independent variable where the data is measured.
				Should usually be an M-length sequence or an (k,M)-shaped array for
				functions with k predictors, but can actually be any object
	y :         array_like
				The dependent data, a length M array - nominally ``f(xdata, ...)``
	p0 :        array_like, optional
				Initial guess for the parameters (length N). If None, then the
				initial values will all be 1 (if the number of parameters for the
				function can be determined using introspection, otherwise a
				ValueError is raised).
	xerr :      array_like, optional
				Determines the uncertainty in `xdata` (used only for plotting).
	yerr :      None or M-length sequence or MxM array, optional
				Determines the uncertainty in `ydata`. If we define residuals as
				``r = ydata - f(xdata, *popt)``, then the interpretation of `sigma`
				depends on its number of dimensions:
					- A 1-D `sigma` should contain values of standard deviations of
					  errors in `ydata`. In this case, the optimized function is
					  ``chisq = sum((r / sigma) ** 2)``.
					- A 2-D `sigma` should contain the covariance matrix of
					  errors in `ydata`. In this case, the optimized function is
					  ``chisq = r.T @ inv(sigma) @ r``.
					- None (default) is equivalent of 1-D `sigma` filled with ones
	plot :      bool, optional
				Set to 'True' if you want to plot the data
	plot_data_and_curve_fit_kw :
	kwargs :    sent to 'scipy.optimize.curve_fit'

	Returns
	-------

	"""
	if plot_data_and_curve_fit_kw is None:
		plot_data_and_curve_fit_kw = {}

	x, xerr = to_numpy(x, xerr)
	y, yerr = to_numpy(y, yerr)
	p0 = val(p0)

	if xerr is not None and xerr.size == 1:
		xerr = np.repeat(xerr, len(x))
	if yerr is not None and yerr.size == 1:
		yerr = np.repeat(yerr, len(y))

	# Curve-Fit
	p_opt, p_cov = scipy.optimize.curve_fit(fit_fcn, x, y, p0=p0, sigma=yerr, **kwargs)

	chi2_red = chi_squared_test(f_exp=fit_fcn(x, *p_opt), f_obs=y, f_obs_err=yerr, ddof=len(p_opt), reduced=True)

	p_opt = from_numpy(p_opt, np.sqrt(np.diag(p_cov)))

	if plot:
		Ax = gr.Axes()
		Ax.plot_data_and_curve_fit(x, y, fit_fcn,
				xerr=xerr, yerr=yerr,
				p_opt=p_opt, p_cov=p_cov,
				**plot_data_and_curve_fit_kw)

		return p_opt, p_cov, chi2_red, Ax

	return p_opt, p_cov, chi2_red


def find_peaks(
		y, yerr=None,
		x=None, xerr=None,
		n_fit_points=None,
		plot=False,
		**kwargs):
	"""

	Parameters
	----------
	x :
	y :
	xerr :
	yerr :
	kwargs :

	Returns
	-------

	"""

	x, xerr = to_numpy(x, xerr)
	y, yerr = to_numpy(y, yerr)
	if x is None:
		x = np.arange(len(y))
	if xerr is not None and xerr.size == 1:
		xerr = np.repeat(xerr, len(x))
	if yerr is not None and yerr.size == 1:
		yerr = np.repeat(yerr, len(y))

	peaks, properties = scipy.signal.find_peaks(val(y), **kwargs)

	if n_fit_points is None:
		x_peaks = x[peaks]
		y_peaks = y[peaks]
		if xerr is not None:
			x_peaks = from_numpy(x_peaks, xerr[peaks])
		if yerr is not None:
			y_peaks = from_numpy(y_peaks, yerr[peaks])

	else:
		# fit quadratic around peaks
		def quadratic(x, a, b, c):
			return a * x ** 2 + b * x + c

		x_peaks = from_numpy(np.zeros(len(peaks)), 0)
		y_peaks = from_numpy(np.zeros(len(peaks)), 0)
		for i in range(len(peaks)):
			# Define a small range around the peak
			left = max(0, peaks[i] - n_fit_points)
			right = min(len(y), peaks[i] + n_fit_points + 1)

		# Perform quadratic fit
		x0 = np.mean(x[left:right])
		popt = curve_fit(fit_fcn=quadratic,
				x=x[left:right] - x0,
				y=y[left:right],
				xerr=xerr[left:right] if xerr is not None else None,
				yerr=yerr[left:right] if yerr is not None else None,
				p0=[0, 0, y[peaks[i]]],
		)[0]
		a_center, b_center, c_center = popt
		a = a_center
		b = b_center - 2 * a_center * x0
		c = a_center * x0 ** 2 - b_center * x0 + c_center

		if val(a) == 0:
			raise ValueError("Quadratic fit failed: a=0")

		x_peaks[i] = - b / (2 * a)  # Vertex of the parabola (peak location)
		y_peaks[i] = quadratic(x_peaks[i], a, b, c)

	if plot:
		Ax = gr.Axes()
		Ax.plot_errorbar(x, y,
				xerr=xerr, yerr=yerr)
		Ax.plot_errorbar(x_peaks, y_peaks,
				marker="o", color="black", linestyle="none")

		return x_peaks, y_peaks, peaks, properties, Ax

	return x_peaks, y_peaks, peaks, properties



import numpy as np
import scipy.special
import scipy.signal
import scipy.interpolate

norm_axis_idx = np.core.multiarray.normalize_axis_index
meshgrid = np.mgrid
gamma = scipy.special.gamma
windows = scipy.signal.windows
sliding_window = np.lib.stride_tricks.sliding_window_view


def array(*args):
	"""
	Convert any data type to numpy.ndarray

	Example:
		>>> x = range(5)  # x = [0,1,2,3,4]
		>>> y = array(x)  # y = np.array([0,1,2,3,4])

	Args:
		*args:          Send as many arrays as you'd like

	Returns:

	"""
	args = list(args)
	for i in range(len(args)):
		if args[i] is not None:
			args[i] = np.asarray(args[i]).squeeze()
	if len(args) == 1:
		return args[0]
	return args


def sigmoid(x: np.ndarray, L: float = 1, k: float = 1, x0: float = 0):
	"""
	Sigmoid/logistic function with parameters L,k,x0 (standard sigmoid has values 1,1,0)

	Args:
		x:          Points to evaluate
		L:          Supremum / maximal asymptotic value
		k:          Steepness
		x0:         Center

	Returns:

	"""
	out = L / (1 + np.exp(-k * (x - x0)))
	return out


def buffer(y: np.ndarray, frame: int, overlap: int) -> np.ndarray:
	"""
	Buffer a 1D array into segments with overlap. Output is a 2D matrix.
	Equivalent to MATLAB's <buffer> function.

	Args:
		y ():               Input array
		frame ():           Length of each segment
		overlap ():         Overlap of every two adjacent segments

	Returns:
		2D buffered array
	"""

	assert y.ndim == 1, "Input array must be 1D"

	hop = frame - overlap
	nframes = int(np.ceil(y.size / hop))

	y_padded = np.concatenate([np.zeros(overlap), y, np.zeros(frame)])
	indices = np.sum(meshgrid[0:frame, 0:nframes * hop:hop], axis=0)
	out = np.take(y_padded, indices, axis=0)
	return out


def unbuffer(y: np.ndarray, frame: int, overlap: int) -> np.ndarray:
	"""
	Perform the inverse operation of the <buffer> function.

	Args:
		y ():               Buffered matrix
		frame ():           Length of each segment
		overlap ():         Overlap of every two adjacent segments

	Returns:
		1D unbuffered array
	"""
	if y.ndim == 1:
		y = np.tile(y, (frame, 1))
	assert y.ndim == 2
	frame = y.shape[0]
	out = y[min(overlap + 1, frame) - 1:, :]
	out = np.reshape(out.T, -1)
	return out


def interp1(y: np.ndarray, n: int, axis: int = -1, *args, **kwargs) -> np.ndarray:
	"""
	1D interpolation or extrapolation.

	Args:
		y:                  Input array
		n:                  Desired length of output array
		axis:               Axis along which to interpolate
		*args:              Passed to scipy.interpolate.interp1d
		**kwargs:           Passed to scipy.interpolate.interp1d

	Returns:

	"""

	if y.shape[axis] == n:
		return y

	h = scipy.interpolate.interp1d(np.linspace(0, n - 1, y.shape[axis]), y, axis=axis, *args, **kwargs)
	out = h(range(n))
	return out


def rms(x: np.ndarray, *args, **kwargs):
	"""
	Root mean squared

	Args:
		x:
		*args:
		**kwargs:

	Returns:

	"""
	return np.sqrt(np.mean(x ** 2, *args, **kwargs))


def rescale(x: np.ndarray, lower=0, upper=1):
	x = x.astype(float)
	# bring to [0, 1]
	x -= np.min(x)
	x /= np.max(x)
	# bring to [lower, upper]
	x *= upper - lower
	x += lower
	return x



# Importing necessary packages
import tkinter as tk
from tkinter import *
from pytube import YouTube
from tkinter import messagebox, filedialog


# Defining CreateWidgets() function
# to create necessary tkinter widgets
def Widgets():
	head_label = Label(root, text="YouTube Video Downloader Using Tkinter",
			padx=15,
			pady=15,
			font="SegoeUI 14",
			bg="palegreen1",
			fg="red")
	head_label.grid(row=1,
			column=1,
			pady=10,
			padx=5,
			columnspan=3)

	link_label = Label(root,
			text="YouTube link :",
			bg="salmon",
			pady=5,
			padx=5)
	link_label.grid(row=2,
			column=0,
			pady=5,
			padx=5)

	root.linkText = Entry(root,
			width=35,
			textvariable=video_Link,
			font="Arial 14")
	root.linkText.grid(row=2,
			column=1,
			pady=5,
			padx=5,
			columnspan=2)

	destination_label = Label(root,
			text="Destination :",
			bg="salmon",
			pady=5,
			padx=9)
	destination_label.grid(row=3,
			column=0,
			pady=5,
			padx=5)

	root.destinationText = Entry(root,
			width=27,
			textvariable=download_Path,
			font="Arial 14")
	root.destinationText.grid(row=3,
			column=1,
			pady=5,
			padx=5)

	browse_B = Button(root,
			text="Browse",
			command=Browse,
			width=10,
			bg="bisque",
			relief=GROOVE)
	browse_B.grid(row=3,
			column=2,
			pady=1,
			padx=1)

	Download_B = Button(root,
			text="Download Video",
			command=Download,
			width=20,
			bg="thistle1",
			pady=10,
			padx=15,
			relief=GROOVE,
			font="Georgia, 13")
	Download_B.grid(row=4,
			column=1,
			pady=20,
			padx=20)


# Defining Browse() to select a
# destination folder to save the video

def Browse():
	# Presenting user with a pop-up for
	# directory selection. initialdir
	# argument is optional Retrieving the
	# user-input destination directory and
	# storing it in downloadDirectory
	download_Directory = filedialog.askdirectory(
			initialdir="YOUR DIRECTORY PATH", title="Save Video")

	# Displaying the directory in the directory
	# textbox
	download_Path.set(download_Directory)


# Defining Download() to download the video


def Download():
	# getting user-input Youtube Link
	Youtube_link = video_Link.get()

	# select the optimal location for
	# saving file's
	download_Folder = download_Path.get()

	# Creating object of YouTube()
	getVideo = YouTube(Youtube_link)

	# Getting all the available streams of the
	# youtube video and selecting the first
	# from the
	videoStream = getVideo.streams.first()

	# Downloading the video to destination
	# directory
	videoStream.download(download_Folder)

	# Displaying the message
	messagebox.showinfo("SUCCESSFULLY",
			"DOWNLOADED AND SAVED IN\n"
			+ download_Folder)


# Creating object of tk class
root = tk.Tk()

# Setting the title, background color
# and size of the tkinter window and
# disabling the resizing property
root.geometry("520x280")
root.resizable(False, False)
root.title("YouTube Video Downloader")
root.config(background="PaleGreen1")

# Creating the tkinter Variables
video_Link = StringVar()
download_Path = StringVar()

# Calling the Widgets() function
Widgets()

# Defining infinite loop to run
# application
root.mainloop()



import numpy as np
from liron_utils import uncertainties_math as un
from liron_utils import graphics


t = np.linspace(0, 10, 101)
x = np.sin(2 * np.pi * 1 * t)
xerr = 0.1 * np.random.uniform(size=t.size)

x = un.from_numpy(x, xerr)

ax = graphics._AxesLiron()
graphics.plot_errorbar(ax.axs, t, x)
ax.ax_props(ax.axs, show_fig=True)
pass



import sys
import atexit
import logging
from logging.handlers import RotatingFileHandler
import time
import inspect
from colorama import Fore, Back, Style

from .dicts import dict_


class Logger:
    NAME2LEVEL = dict_(logging._nameToLevel)
    LEVEL2NAME_OLD = dict_(logging._levelToName)
    LEVEL2NAME = dict_({
        logging.CRITICAL: 'Fatal',
        logging.ERROR:    'Error',
        logging.WARNING:  'Warn',
        logging.INFO:     'Info',
        logging.DEBUG:    'Debug',
        logging.NOTSET:   'Notset',
    })
    ENTER_MSG = "STARTING LOGGER\t------------------------------------"
    EXIT_MSG = "CLOSING LOGGER\t------------------------------------"

    ENTER_CALLED = False

    def __init__(self,
            file_name: str = "./logs.log",
            min_level_file: int = logging.INFO,
            min_level_console: int = None,
            default_level: int = logging.INFO,
            log_message_format: str = "%(asctime)s | %(levelname)-14s | %(filename)-15s:%(lineno)4d | %(class_func)-30s >> %(message)s",
            max_file_size: int = 10 * 1024 * 1024,
            backup_count: int = 5,
    ):
        """
        Initialize logger.

        Examples
        --------
        >>> with Logger(min_level_console=Logger.NAME2LEVEL["WARNING"]) as logger:
        >>>      logger.log("Hello")  # Will not be printed
        >>>      logger.log("World", logging.ERROR)  # Will be printed

        >>> logger = Logger()  # __enter__ is called automatically
        >>> logger.log("Hello")  # __exit__ will be called automatically upon code completion

        >>> logger = Logger()
        >>> raise ValueError("This is an error.")  # Uncaught exception will be logged and the logger will be closed

        Parameters
        ----------
        file_name :                 str, optional
            Logger file path. Default is "./logs.log"
        min_level_file              str, optional
            Level of messages to print to file. Default is INFO
        min_level_console :         int, optional
            Level of messages to print to console. If None, will not print
        default_level :             int, optional
            Default logging level, if no level is provided. Default is INFO
        log_message_format :        str, optional
            Default string format of a message
        max_file_size :             int, optional
            Maximum size of log file in bytes before rotation. Default is 10MB
        backup_count :              int, optional
            Number of backup files to keep. Default is 5
        """
        if not file_name.endswith(".log"):
            file_name += ".log"

        with open(file_name, 'a'):  # Create the file if it doesn't exist
            pass
        self.logger = logging.getLogger(file_name)
        self.default_level = default_level
        self.logger.setLevel(logging.DEBUG)

        # Create a file handler to write log messages to a file
        class _Formatter(logging.Formatter):
            # Formatter that changes the level name to a shorthand version (e.g., WARNING->Warn).
            def format(self, record):
                record.levelname = record.levelname.replace(Logger.LEVEL2NAME_OLD[record.levelno],
                        Logger.LEVEL2NAME[record.levelno])

                record.class_func = self._get_class_func(record)

                return super().format(record)

            @staticmethod
            def _get_class_func(record):
                frame = inspect.currentframe()
                while frame:
                    if frame.f_code.co_name == record.funcName:
                        # Look for 'self' or 'cls' in the frame's local variables
                        local_self = frame.f_locals.get('self', None) or frame.f_locals.get('cls', None)
                        if local_self:
                            return f'{local_self.__class__.__name__}.{record.funcName}'
                    frame = frame.f_back
                return record.funcName  # Fallback

        # Use RotatingFileHandler instead of FileHandler for file size rotation
        self.file_handler = RotatingFileHandler(
                file_name,
                maxBytes=max_file_size,
                backupCount=backup_count
        )
        if min_level_file is not None:
            self.file_handler.setLevel(min_level_file)
            formatter = _Formatter(log_message_format)
            self.file_handler.setFormatter(formatter)
            self.logger.addHandler(self.file_handler)

        # Create a console handler to write log messages to the console
        class _FormatterColored(_Formatter):
            # Formatter that adds colors to log messages based on their level.
            COLORS = {
                logging.CRITICAL: Fore.LIGHTWHITE_EX + Back.LIGHTRED_EX,
                logging.ERROR:    Fore.RED,
                logging.WARNING:  Fore.YELLOW,
                logging.INFO:     Fore.BLUE,
                logging.DEBUG:    Fore.MAGENTA,
                logging.NOTSET:   Fore.BLACK
            }

            def format(self, record):
                color = self.COLORS.get(record.levelno, "")
                record.levelname = f"{color}{record.levelname}{Style.RESET_ALL}"
                return super().format(record)

        self.console_handler = logging.StreamHandler(sys.stdout)
        if min_level_console is not None:
            self.console_handler.setLevel(min_level_console)
            formatter = _FormatterColored(log_message_format)
            self.console_handler.setFormatter(formatter)
            self.logger.addHandler(self.console_handler)

        atexit.register(self.__exit__, None, None, None)  # Register exit function

        excepthook_old = sys.excepthook  # Save the old exception hook

        def excepthook_new(exc_type, exc_value, exc_traceback):
            self.__exit__(exc_type, exc_value, exc_traceback)
            excepthook_old(exc_type, exc_value, exc_traceback)

        sys.excepthook = excepthook_new  # Set the new exception hook

        self.__enter__()

    def __enter__(self):
        if self.ENTER_CALLED:
            return self

        self.info(self.ENTER_MSG)
        self.ENTER_CALLED = True
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            if not issubclass(exc_type, KeyboardInterrupt):  # Ignore keyboard interrupts
                self.error("Uncaught exception:", exc_info=(exc_type, exc_value, exc_traceback))

        self.info(self.EXIT_MSG)
        self.logger.removeHandler(self.file_handler)
        self.file_handler.close()
        self.logger.removeHandler(self.console_handler)
        self.console_handler.close()

    def log(self,
            msg: str,
            level: int = None,
            exc_info: bool = False,
            time_log: float = None,
            f: str = ':.3f',
            stacklevel: int = 2,
            *args, **kwargs):
        """
        Send a log message to the log file.

        Parameters
        ----------
        msg :               str
            Message string
        level :             int
            Logging level (one of self.LEVELS)
        exc_info :          bool
            Error info
        time_log :          float
            Time log
        f :                 str
            Format string for time log
        stacklevel :        int
            Stack level for the log message
        *args, **kwargs :   Passed to Logger.log

        **kwargs :      dict
            Passed to Logger.log

        Returns
        -------

        """

        if level is None:
            level = self.default_level

        kwargs = dict(stacklevel=stacklevel) | kwargs

        tm = time.time()
        if time_log is not None:
            msg = msg.replace('{}', '{' + f + '}')
            msg = msg.format(tm - time_log)

        self.logger.log(
                level=level,
                msg=msg,
                exc_info=exc_info,
                *args, **kwargs)

        return tm

    def debug(self, msg: str, *args, **kwargs):
        kwargs = dict(stacklevel=3) | kwargs
        return self.log(msg, level=logging.DEBUG, *args, **kwargs)

    def info(self, msg: str, *args, **kwargs):
        kwargs = dict(stacklevel=3) | kwargs
        return self.log(msg, level=logging.INFO, *args, **kwargs)

    def warning(self, msg: str, *args, **kwargs):
        kwargs = dict(stacklevel=3) | kwargs
        return self.log(msg, level=logging.WARNING, *args, **kwargs)

    warn = warning

    def error(self, msg: str, *args, **kwargs):
        kwargs = dict(stacklevel=3) | kwargs
        return self.log(msg, level=logging.ERROR, *args, **kwargs)

    def critical(self, msg: str, *args, **kwargs):
        kwargs = dict(stacklevel=3) | kwargs
        return self.log(msg, level=logging.CRITICAL, *args, **kwargs)

    fatal = critical



"""
Color hexadecimal values
"""

WHITE: str = "#FFFFFF"
GRAY_A: str = "#DDDDDD"
GRAY_B: str = "#BBBBBB"
GRAY_C: str = "#888888"
GRAY_D: str = "#444444"
GRAY_E: str = "#222222"
BLACK: str = "#000000"
LIGHTER_GRAY: str = "#DDDDDD"
LIGHTER_GREY: str = "#DDDDDD"
LIGHT_GRAY: str = "#BBBBBB"
LIGHT_GREY: str = "#BBBBBB"
GRAY: str = "#888888"
GREY: str = "#888888"
DARK_GRAY: str = "#444444"
DARK_GREY: str = "#444444"
DARKER_GRAY: str = "#222222"
DARKER_GREY: str = "#222222"
BLUE_A: str = "#C7E9F1"
BLUE_B: str = "#9CDCEB"
BLUE_C: str = "#58C4DD"
BLUE_D: str = "#29ABCA"
BLUE_E: str = "#236B8E"
PURE_BLUE: str = "#0000FF"
BLUE: str = "#58C4DD"
DARK_BLUE: str = "#236B8E"
TEAL_A: str = "#ACEAD7"
TEAL_B: str = "#76DDC0"
TEAL_C: str = "#5CD0B3"
TEAL_D: str = "#55C1A7"
TEAL_E: str = "#49A88F"
TEAL: str = "#5CD0B3"
GREEN_A: str = "#C9E2AE"
GREEN_B: str = "#A6CF8C"
GREEN_C: str = "#83C167"
GREEN_D: str = "#77B05D"
GREEN_E: str = "#699C52"
PURE_GREEN: str = "#00FF00"
GREEN: str = "#83C167"
YELLOW_A: str = "#FFF1B6"
YELLOW_B: str = "#FFEA94"
YELLOW_C: str = "#FFFF00"
YELLOW_D: str = "#F4D345"
YELLOW_E: str = "#E8C11C"
YELLOW: str = "#FFFF00"
GOLD_A: str = "#F7C797"
GOLD_B: str = "#F9B775"
GOLD_C: str = "#F0AC5F"
GOLD_D: str = "#E1A158"
GOLD_E: str = "#C78D46"
GOLD: str = "#F0AC5F"
RED_A: str = "#F7A1A3"
RED_B: str = "#FF8080"
RED_C: str = "#FC6255"
RED_D: str = "#E65A4C"
RED_E: str = "#CF5044"
PURE_RED: str = "#FF0000"
RED: str = "#FC6255"
MAROON_A: str = "#ECABC1"
MAROON_B: str = "#EC92AB"
MAROON_C: str = "#C55F73"
MAROON_D: str = "#A24D61"
MAROON_E: str = "#94424F"
MAROON: str = "#C55F73"
PURPLE_A: str = "#CAA3E8"
PURPLE_B: str = "#B189C6"
PURPLE_C: str = "#9A72AC"
PURPLE_D: str = "#715582"
PURPLE_E: str = "#644172"
PURPLE: str = "#9A72AC"
PINK: str = "#D147BD"
LIGHT_PINK: str = "#DC75CD"
ORANGE: str = "#FF862F"
LIGHT_BROWN: str = "#CD853F"
DARK_BROWN: str = "#8B4513"
GRAY_BROWN: str = "#736357"
GREY_BROWN: str = "#736357"

# %% Custom Colors
ORANGE_A = "#FF7F0E"
ORANGE_B = "#D95319"
PINK_A = "#E377C2"

__all__ = [item for item in dir() if not item.startswith('_')]

# %% Custom Variables
RGB = (RED_A, GREEN_A, BLUE_A)



import os
import functools
import warnings
import copy
from typing import Iterable
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes as Axes_plt
from matplotlib.ticker import ScalarFormatter
import matplotlib.animation

from .utils.default_kwargs import merge_kwargs
from ..time import TIME_STR, get_time_str
from ..files import MAIN_FILE_DIR, open_file, mkdirs
from ..pure_python.dicts import DL_to_LD


# TODO: change Ax.axs to be 1D (top->down, left->right), enable custom shapes (large, small axes)


class _Axes:
    def __init__(self,
            shape: tuple[int] = (1, 1),
            grid_layout: list[list[tuple]] = None,
            sharex: bool | str = False, sharey: bool | str = False,
            projection: str = None,
            layout: str = None,
            fig: Figure = None, axs: Axes_plt | Iterable[Axes_plt] = None,
            subplot_kw: dict = None, gridspec_kw: dict = None, **fig_kw):
        """
        Create a new figure with (possibly) subplots using the plt.subplots() function.

        Parameters
        ----------
        shape :             tuple[int], default: (1, 1)
                            number of rows, columns (in case of subplots).

        layout :            list[list[tuple]], optional
                            ([<row_start>, <row_end>], [<col_start>, <col_end>]),
                            e.g., [[(0, 2), (0, 2)], [(0, 1), (2, 4)]]

        sharex, sharey :    bool or {'none', 'all', 'row', 'col'}, default: False
                            Share the x or y `~matplotlib.axis` with sharex and/or sharey.
                            The axis will have the same limits, ticks, and scale as the axis
                            of the shared axes.

        projection :        {None, 'aitoff', 'hammer', 'lambert', 'mollweide', 'polar', 'rectilinear', str}, default: None
                            The projection type of the subplot (`~.axes.Axes`). *str* is the
                            name of a custom projection, see `~matplotlib.projections`. The
                            default None results in a 'rectilinear' projection.

        layout :            {'constrained', 'compressed', 'tight', 'none', `.LayoutEngine`, None}, default: None
                            The layout mechanism for positioning of plot elements to avoid
                            overlapping Axes decorations (labels, ticks, etc). Note that layout
                            managers can measurably slow down figure display.

                            - 'constrained': The constrained layout solver adjusts axes sizes
                              to avoid overlapping axes decorations.  Can handle complex plot
                              layouts and colorbars, and is thus recommended.

                              See :ref:`constrainedlayout_guide`
                              for examples.

                            - 'compressed': uses the same algorithm as 'constrained', but
                              removes extra space between fixed-aspect-ratio Axes.  Best for
                              simple grids of axes.

                            - 'tight': Use the tight layout mechanism. This is a relatively
                              simple algorithm that adjusts the subplot parameters so that
                              decorations do not overlap. See `.Figure.set_tight_layout` for
                              further details.

                            - 'none': Do not use a layout engine.

                            - A `.LayoutEngine` instance. Builtin layout classes are
                              `.ConstrainedLayoutEngine` and `.TightLayoutEngine`, more easily
                              accessible by 'constrained' and 'tight'.  Passing an instance
                              allows third parties to provide their own layout engine.

                            If not given, fall back to using the parameters *tight_layout* and
                            *constrained_layout*, including their config defaults
                            :rc:`figure.autolayout` and :rc:`figure.constrained_layout.use`.

        fig :               Figure, optional
                            Usually None, or send an existing figure

        axs :               Axes, optional
                            Usually None, or send own axis/axes

        Other Parameters
        ----------------
        subplot_kw :        dict, optional
                            Not interesting

        gridspec_kw :       dict, optional
                            left, right, top, bottom : float, optional
                                Extent of the subplots as a fraction of figure width or height.
                                Left cannot be larger than right, and bottom cannot be larger than
                                top. If not given, the values will be inferred from a figure or
                                rcParams at draw time. See also `GridSpec.get_subplot_params`.

                            wspace : float, optional
                                The amount of width reserved for space between subplots,
                                expressed as a fraction of the average axis width.
                                If not given, the values will be inferred from a figure or
                                rcParams when necessary. See also `GridSpec.get_subplot_params`.

                            hspace : float, optional
                                The amount of height reserved for space between subplots,
                                expressed as a fraction of the average axis height.
                                If not given, the values will be inferred from a figure or
                                rcParams when necessary. See also `GridSpec.get_subplot_params`.

                            width_ratios : array-like of length *ncols*, optional
                                Defines the relative widths of the columns. Each column gets a
                                relative width of ``width_ratios[i] / sum(width_ratios)``.
                                If not given, all columns will have the same width.

                            height_ratios : array-like of length *nrows*, optional
                                Defines the relative heights of the rows. Each row gets a
                                relative height of ``height_ratios[i] / sum(height_ratios)``.
                                If not given, all rows will have the same height.

        fig_kw :            dict, optional
                            num : int or str or `.Figure` or `.SubFigure`, optional
                                A unique identifier for the figure.

                                If a figure with that identifier already exists, this figure is made
                                active and returned. An integer refers to the ``Figure.number``
                                attribute, a string refers to the figure label.

                                If there is no figure with the identifier or *num* is not given, a new
                                figure is created, made active and returned.  If *num* is an int, it
                                will be used for the ``Figure.number`` attribute, otherwise, an
                                auto-generated integer value is used (starting at 1 and incremented
                                for each new figure). If *num* is a string, the figure label and the
                                window title is set to this value.  If num is a ``SubFigure``, its
                                parent ``Figure`` is activated.

                            figsize : (float, float), default: :rc:`figure.figsize`
                                Width, height in inches.

                            dpi : float, default: :rc:`figure.dpi`
                                The resolution of the figure in dots-per-inch.

                            facecolor : color, default: :rc:`figure.facecolor`
                                The background color.

                            edgecolor : color, default: :rc:`figure.edgecolor`
                                The border color.

                            frameon : bool, default: True
                                If False, suppress drawing the figure frame.

                            FigureClass : subclass of `~matplotlib.figure.Figure`
                                If set, an instance of this subclass will be created, rather than a
                                plain `.Figure`.

                            clear : bool, default: False
                                If True and the figure already exists, then it is cleared.

                            **kwargs
                                Additional keyword arguments are passed to the `.Figure` constructor.


        Examples
        --------
            >> from plotting import plot
            >> Ax = AxesLiron([2, 3])
            >> t = np.linspace(0, 10, 1001)
            >> plot(Ax.axs[0,0], t, np.sin(t))
            >> plot(Ax.axs[1,0], t, np.cos(t))
            >> plot(Ax.axs[0,1], t, np.log(t))
            >> plot(Ax.axs[1,1], t, np.exp(t))
            >> plot(Ax.axs[0,2], t, np.sqrt(t))
            >> plot(Ax.axs[1,2], t, np.square(t))

            >> Ax.set_props(sup_title="abc", ax_title=[1, 2, 3, 4, 5, 6], grid=[True, True, False, True, False, True], show_fig=True)

        """

        self.fig = fig
        self.axs = np.atleast_2d(axs)

        if fig is None and axs is None:
            if subplot_kw is None:
                subplot_kw = dict()
            subplot_kw = {"projection": projection} | subplot_kw
            fig_kw = {"layout": layout} | fig_kw
            fig_kw = merge_kwargs(fig_kw=fig_kw)["fig_kw"]
            padding = [None] * 4
            for i, key in enumerate(["w_pad", "h_pad", "wspace", "hspace"]):
                if key in fig_kw:
                    padding[i] = fig_kw.pop(key)

            nrows, ncols = shape

            if grid_layout is None:
                self.fig, self.axs = plt.subplots(nrows=nrows, ncols=ncols,
                        sharex=sharex, sharey=sharey,
                        squeeze=False,
                        subplot_kw=subplot_kw, gridspec_kw=gridspec_kw, **fig_kw)
            else:
                if gridspec_kw is None:
                    gridspec_kw = dict()
                if sharex == "all":
                    sharex = True
                elif sharex == "none":
                    sharex = False
                if sharey == "all":
                    sharey = True
                elif sharey == "none":
                    sharey = False
                assert type(sharex) is bool and type(sharey) is bool, "'sharex' and 'sharey' must be bool."

                self.fig = plt.figure(**fig_kw)
                gs = self.fig.add_gridspec(nrows=nrows, ncols=ncols, **gridspec_kw)

                ax_share = None
                axs = np.ones(shape=shape, dtype=Axes_plt)
                for i, (row_range, col_range) in enumerate(grid_layout):
                    if isinstance(row_range, int):
                        row_range = (row_range, row_range + 1)
                    if isinstance(col_range, int):
                        col_range = (col_range, col_range + 1)
                    r0, r1 = row_range
                    c0, c1 = col_range

                    axs[r0:r1, c0:c1] = 0
                    axs[r0, c0] = self.fig.add_subplot(gs[r0:r1, c0:c1],
                            sharex=ax_share if sharex else None,
                            sharey=ax_share if sharey else None,
                            **subplot_kw
                    )

                    if ax_share is None:
                        ax_share = axs[r0, c0]

                for (r, c) in np.argwhere(axs == 1):
                    axs[r, c] = self.fig.add_subplot(gs[r, c],
                            sharex=ax_share if sharex else None,
                            sharey=ax_share if sharey else None,
                            **subplot_kw
                    )

                self.axs = axs

            from matplotlib.layout_engine import ConstrainedLayoutEngine
            if type(self.fig.get_layout_engine()) is ConstrainedLayoutEngine:
                self.fig.get_layout_engine().set(w_pad=padding[0], h_pad=padding[1], hspace=padding[2],
                        wspace=padding[3])

        elif fig is not None:
            self.axs = np.atleast_2d(self.fig.axes)

        elif axs is not None:
            self.fig = self.axs[0, 0].figure

        self.func_animation = None

    def __getitem__(self, item):
        out = copy.deepcopy(self)
        out.axs = np.atleast_2d(self.axs[item])
        return out

    @staticmethod
    def _vectorize(cls, ax: Axes_plt = None, **vec_params):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if ax is not None:
                    return func(ax, *args, **vec_params, **kwargs)

                # Vectorize the function
                m, n = cls.axs.shape

                params_list = DL_to_LD(vec_params)
                if params_list is None or len(params_list) != m * n:
                    params_list = np.repeat(vec_params, m * n)

                out = np.empty((m, n), dtype=object)

                for i in range(m):
                    for j in range(n):
                        out[i, j] = func(cls.axs[i, j], *args, **params_list[j * m + i], **kwargs)

                return out

            return wrapper

        return decorator

    @staticmethod
    def _merge_kwargs(key, **kwargs):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args):
                kwargs_merged = merge_kwargs(**{key: kwargs})[key]

                return func(*args, **kwargs_merged)

            return wrapper

        return decorator

    def draw_xy_lines(self, **xy_lines_kw):
        @self._merge_kwargs("xy_lines_kw", **xy_lines_kw)
        @self._vectorize(cls=self)
        def _draw_xy_lines(ax: Axes_plt, **xy_lines_kw):
            """
            Draw x-y axes lines to look bolder than the rest of the grid lines

            Args:
                ax:
                **axis_lines_kw:

            Returns:

            """

            if hasattr(ax, "axis_lines_drawn") or hasattr(ax, 'zaxis') or len(
                    ax.images) > 0:  # Don't draw axis lines for 3D plots and images
                return

            xlim = ax.get_xlim()
            ylim = ax.get_ylim()

            ax.axhline(**xy_lines_kw)
            ax.axvline(**xy_lines_kw)

            ax.set_xlim(*xlim, auto=True)
            ax.set_ylim(*ylim, auto=True)

            ax.axis_lines_drawn = True

        _draw_xy_lines()

    def sup_title(self, title: str):
        # todo: for unknown reason, color is not automatically inherited from text.color
        self.fig.suptitle(title, color=plt.rcParams["text.color"])

    def ax_axis(self, axis: bool | str):
        """

            Parameters
            ----------
            ax :
            axis :  ================ ===========================================================
                    Value            Description
                    ================ ===========================================================
                    'off' or `False` Hide all axis decorations, i.e. axis labels, spines,
                                     tick marks, tick labels, and grid lines.
                                     This is the same as `~.Axes.set_axis_off()`.
                    'on' or `True`   Do not hide all axis decorations, i.e. axis labels, spines,
                                     tick marks, tick labels, and grid lines.
                                     This is the same as `~.Axes.set_axis_on()`.
                    'equal'          Set equal scaling (i.e., make circles circular) by
                                     changing the axis limits. This is the same as
                                     ``ax.set_aspect('equal', adjustable='datalim')``.
                                     Explicit data limits may not be respected in this case.
                    'scaled'         Set equal scaling (i.e., make circles circular) by
                                     changing dimensions of the plot box. This is the same as
                                     ``ax.set_aspect('equal', adjustable='box', anchor='C')``.
                                     Additionally, further autoscaling will be disabled.
                    'tight'          Set limits just large enough to show all data, then
                                     disable further autoscaling.
                    'auto'           Automatic scaling (fill plot box with data).
                    'image'          'scaled' with axis limits equal to data limits.
                    'square'         Square plot; similar to 'scaled', but initially forcing
                                     ``xmax-xmin == ymax-ymin``.
                    ================ ===========================================================

            Returns
            -------

            """

        @self._vectorize(cls=self, axis=axis)
        def _ax_axis(ax: Axes_plt, axis: bool | str):
            ax.axis(axis)

        _ax_axis()

    def ax_spines(self, spines: str | list | bool):
        """
        Show axis spines (boundaries)

        Parameters
        ----------
        spines :        str | list | bool

        """

        @self._vectorize(cls=self, spines=spines)
        def _ax_spines(ax: Axes_plt, spines: str | list | bool):
            locs = np.array(["left", "bottom", "top", "right"])

            if type(spines) is str:
                spines = [spines]

            if type(spines) is list:
                locs = np.array(spines)
                idx = [True] * locs.size
            elif type(spines) is bool:
                idx = [False] * locs.size
                if spines:
                    idx = [True] * locs.size
            else:
                raise ValueError("'spines' must be one of (str, list, bool).")

            if locs[idx].size > 0:
                ax.spines[locs[idx].tolist()].set_visible(True)
            if locs[np.logical_not(idx)].size > 0:
                ax.spines[locs[np.logical_not(idx)].tolist()].set_visible(False)

        _ax_spines()

    def ax_ticks(self, ticks: bool | list[list] | dict | list[dict], labels: bool | list[list]):
        @self._vectorize(cls=self, ticks=ticks, labels=labels)
        def _ax_ticks(ax: Axes_plt, ticks: bool | list[list] | dict | list[dict], labels: bool | list[list]):
            # todo: if labels is False
            if len(ax.get_shared_x_axes().get_siblings(ax)) > 1 or len(ax.get_shared_y_axes().get_siblings(ax)) > 1:
                warnings.warn(
                        f"Axis shares x or y axis with siblings. Changing ticks would affect all siblings.")  # todo: fix in future

            ndim = 3 if hasattr(ax, "get_zticks") else 2

            if ticks is None or ticks is True:  # show ticks for all axes
                ticks = [True] * ndim
                if labels is None:  # show labels for all axes
                    labels = [True] * ndim
            elif ticks is False:  # hide ticks for all axes
                ticks = [False] * ndim
                if labels is None:  # hide labels for all axes
                    labels = [False] * ndim
            elif type(ticks) is dict:  # only x-axis
                ticks = [ticks] + [True] * (ndim - 1)
            # labels check is done later

            if labels is True:  # show labels for all axes
                labels = [True] * ndim
            elif labels is False:  # hide labels for all axes
                labels = [False] * ndim
            elif labels is None:
                labels = [None] * ndim

            # ticks = np.atleast_1d(ticks)
            # labels = np.atleast_1d(labels)
            assert len(ticks) <= ndim, "len(ticks) must be <= graph dimensionality."
            assert len(labels) <= ndim, "len(labels) must be <= graph dimensionality."

            d = {
                "ticks":  {
                    0: ax.get_xticks(),
                    1: ax.get_yticks(),
                    2: ax.get_zticks() if ndim == 3 else [],
                },

                "labels": {
                    0: ax.get_xticklabels(),
                    1: ax.get_yticklabels(),
                    2: ax.get_zticklabels() if ndim == 3 else [],
                },
                # "offset": {
                # 	0: ax.xaxis.get_offset_text(),
                # 	1: ax.yaxis.get_offset_text(),
                # 	2: ax.zaxis.get_offset_text() if ndim == 3 else [],
                # },
            }

            funcs = [ax.get_xlim, ax.get_ylim]
            if ndim == 3:
                funcs += [ax.get_zlim]

            for i, func in enumerate(funcs):
                lim = func()
                idx = np.logical_and(min(lim) <= d["ticks"][i], d["ticks"][i] <= max(lim))
                if len(d["ticks"][i]) > 0:
                    d["ticks"][i] = d["ticks"][i][idx]
                if len(d["labels"][i]) > 0:
                    d["labels"][i] = np.array(d["labels"][i])[idx]

            for i in range(len(ticks)):  # x,y,z axes
                if ticks[i] is None or ticks[i] is True or ticks[i] is np.True_:  # show ticks
                    pass

                elif ticks[i] is False or ticks[i] is np.False_:  # hide ticks
                    d["ticks"][i] = []
                    if labels[i] is None:  # hide labels
                        d["labels"][i] = []
                # d["offset"][i] = None

                elif type(ticks[i]) is dict:  # custom ticks
                    assert labels[i] is None or labels[
                        i] is np.True_, "When 'ticks' is a dict, 'labels' must not be given."
                    d["ticks"][i] = ticks[i].keys()
                    d["labels"][i] = ticks[i].values()

                elif type(ticks[i]) is list or type(ticks[i]) is np.ndarray:  # custom ticks
                    d["ticks"][i] = ticks[i]
                    d["labels"][i] = ticks[i]
                # d["offset"][i] = None  # todo: check

                else:
                    raise ValueError("'ticks' must be given either as a boolean, list[list] or list[dict].")

                if labels[i] is None or labels[i] is True or labels[i] is np.True_:  # show labels
                    pass

                elif labels[i] is False or labels[i] is np.False_:  # hide labels
                    d["labels"][i] = []
                # d["offset"][i] = None

                elif type(labels[i]) is list or type(labels[i]) is np.ndarray:  # custom labels
                    assert len(labels[i]) == len(ticks[i]), "len(labels[i]) must be equal to len(ticks[i])."
                    d["labels"][i] = labels[i]
                # d["offset"][i] = None  # todo: check

                else:
                    raise ValueError("'labels' must be given either as a boolean or list[list].")

            # Set ticks and labels
            ax.set_xticks(d["ticks"][0], d["labels"][0])
            ax.xaxis.set_major_formatter(
                    ScalarFormatter(useOffset=True))  # todo: returns the tick labels for unknown reason

            ax.set_yticks(d["ticks"][1], d["labels"][1])
            ax.yaxis.set_major_formatter(ScalarFormatter(useOffset=True))

            if ndim == 3:
                ax.set_zticks(d["ticks"][2], d["labels"][2])
                ax.zaxis.set_major_formatter(ScalarFormatter(useOffset=True))

        _ax_ticks()

    def ax_title(self, title: str):
        @self._vectorize(cls=self, title=title)
        def _ax_title(ax: Axes_plt, title: str):
            ax.set_title(title)

        _ax_title()

    def ax_labels(self, labels: list[str]):
        @self._vectorize(cls=self, labels=labels)
        def _ax_labels(ax: Axes_plt, labels: list[str]):
            labels = np.atleast_1d(labels)  # In case labels=None or labels is just 1 list (only xlabel)
            ax.set_xlabel(labels[0])
            if np.size(labels) >= 2:
                ax.set_ylabel(labels[1])
            if np.size(labels) >= 3:
                ax.set_zlabel(labels[2])

        _ax_labels()

    def ax_limits(self, limits: list[float]):
        @self._vectorize(cls=self, limits=limits)
        def _ax_limits(ax: Axes_plt, limits: list[float]):
            limits = np.array(limits, dtype=object)
            limits = np.atleast_2d(limits)  # In case limits=None or limits is just 1 list (only xlim)

            ax.set_xlim(limits[0])
            if len(limits) >= 2:
                ax.set_ylim(limits[1])
            if len(limits) >= 3:
                ax.set_zlim(limits[2])

        _ax_limits()

    def ax_view(self, view: list):
        @self._vectorize(cls=self, view=view)
        def _ax_view(ax: Axes_plt, view: list):
            ax.view_init(view[0], view[1])

        _ax_view()

    def ax_grid(self, grid: bool):
        @self._vectorize(cls=self, grid=grid)
        def _ax_grid(ax: Axes_plt, grid: bool):
            ax.grid(grid)

        _ax_grid()

    def ax_legend(self, legend: bool | list, legend_loc: str):
        @self._vectorize(cls=self, legend=legend, legend_loc=legend_loc)
        def _ax_legend(ax: Axes_plt, legend: bool | list, legend_loc: str):
            if legend is False:
                legend = ax.get_legend()
                if legend is not None:
                    legend.remove()

            elif legend is True:
                handles, labels = ax.get_legend_handles_labels()
                if len(handles) > 0:
                    ax.legend(handles, labels, loc=legend_loc)  # todo [::-1] ?

            else:
                ax.legend(legend, loc=legend_loc)

        _ax_legend()

    def ax_colorbar(self, axs: bool | list[Axes_plt] | list[list[Axes_plt]], **colorbar_kw):
        if axs is False:
            return
        elif axs is True:
            axs = [[ax] for ax in self.fig.axes if len(ax.images) > 0]
        elif type(axs) is Axes_plt:
            axs = [axs]
        elif type(axs) is np.ndarray:
            axs = axs.tolist()

        # convert to a list of lists
        for i in range(len(axs)):  # run through common axes
            if type(axs[i]) is Axes_plt:  # if not a list, convert to a list
                axs[i] = [axs[i]]

            assert type(axs[i]) is list, "'axs' must be a list of lists."
            for j in range(len(axs[i])):
                assert len(axs[i][j].images) > 0, "At least one of the axes provided doesn't have an image plotted."

        # draw colorbar
        for ax_common in axs:
            mappable = ax_common[0].images[0]
            self.fig.colorbar(mappable=mappable, ax=ax_common, **colorbar_kw)

    def ax_face_color(self, face_color):
        @self._vectorize(cls=self, face_color=face_color)
        def _ax_face_color(ax, face_color):
            ax.set_facecolor(face_color)

        _ax_face_color()

    def save_fig(self, file_name: str = None, **savefig_kw):
        """
        Save figure as file

        Args:
            file_name:              File name
            **savefig_kw:           kwargs for function fig.savefig

        Returns:

        """

        file_name = get_savefig_file_name(file_name, mkdir=True)

        format = os.path.splitext(file_name)[-1]

        if format == ".gif":
            assert isinstance(self.func_animation, matplotlib.animation.FuncAnimation), \
                "Call plot_animation() before saving gif."
            self.func_animation.save(file_name, **savefig_kw)
        else:
            self.fig.savefig(file_name, **savefig_kw)

        return file_name

    def show_fig(self):
        self.fig.show()

    def show(self):
        self.show_fig()

    def set_props(self,
            save_file_name: str | bool = False,
            colorbar_kw: dict = None,
            xy_lines_kw: dict = None,
            save_fig_kw: dict = None,
            **set_props_kw):
        """
        Set figure properties after plotting.

        Parameters
        ----------
        save_file_name :    str/bool
                            The file name to be saved.
                            False - don't save, True - save to MAIN_FILE_DIR, str - specify location

        save_fig_kw :       dict, optional
                            Sent to fig.savefig()

        colorbar_kw :       dict, optional
                            Sent to fig.colorbar()

        xy_lines_kw :       dict, optional
                            Sent to ax.axhline()


        Other Parameters
        ----------------
        For the following parameters, a '*' symbol means that it can be vectorized
        to all axes by being sent as a list.

        sup_title :         str, default: None
                            supreme figure title

        *ax_title :         str, default: None
                            axis title

        *axis :             str/bool, default: None

        *spines :           bool/list/str, default: None
                            show axis boundaries.
                            'True' will plot all boundaries.
                            Can be specified as one (or a list) of the strings ["left", "bottom", "top", "right"]

        *ticks :            bool/list, default: None
                            axis ticks
                            'False' will remove all ticks and tick labels.
                            A list will set the ticks for the x,y,z axes

        *tick_labels :     bool/list, default: None
                            axis tick labels
                            'False' will remove all tick labels.
                            A list will set the tick labels for the x,y,z axes

        *labels :           list[str], default: None
                            axis labels for the x,y,z axes

        *limits :           list[float], default: None
                            axis limits for the x,y,z axes, given as a 2-tuple,
                            e.g., limits=[[0,2], [-1,1]] will set xlim=[0,2],ylim=[-1,1]

        *view :             list[float], default: None
                            axis view in case of a 3D plot, given as a 2-tuple.

        *grid :             bool, default: None
                            Show grid.
                            'None' will toggle the grid on and off

        *legend :           bool/list[str], default: True
                            Show legend.
                            A list of strings would set the legend labels

        *legend_loc :       str, default: None
                            The location of the legend:
                            - 'upper left', 'upper right', 'lower left', 'lower right' place the legend at the
                              corresponding corner of the axes.
                            - 'upper center', 'lower center', 'center left', 'center right' place the legend
                              at the center of the corresponding edge of the axes.
                            - 'center' places the legend at the center of the axes.
                            - 'best' places the legend at the location, among the nine locations defined so far, with
                              the minimum overlap with other drawn artists.
                              This option can be quite slow for plots with
                              large amounts of data; your plotting speed may benefit from providing a specific location.
                            The location can also be a 2-tuple giving the coordinates of the lower-left corner of the
                            legend in axes coordinates (in which case bbox_to_anchor will be ignored).

        colorbar :          bool/list[list[Axes], Axes], default: False
                            A boolean decides whether to add a colorbar to each axis containing an image.
                            Otherwise, provide a list of the desired axes.
                            A list inside the list would create a common colorbar.

        *xy_lines :         bool, default: True
                            Draw x-y axis lines (the lines x=0 and y=0).

        *face_color :       color, default: None
                            axes face color

        show_fig :          bool, default: True
                            Show figure at the end of the function run

        open_dir :          bool, default: False
                            If file saved, open directory afterward.

        Returns
        -------

        """

        if colorbar_kw is None:
            colorbar_kw = dict()
        if xy_lines_kw is None:
            xy_lines_kw = dict()
        if save_fig_kw is None:
            save_fig_kw = dict()

        set_props_kw = merge_kwargs(set_props_kw=set_props_kw)["set_props_kw"]

        def caller(func, *args, **kwargs):
            if args[0] is None:
                return None
            return func(*args, **kwargs)

        # Supreme Title
        caller(self.sup_title, set_props_kw["sup_title"])

        # Axes Title
        caller(self.ax_title, set_props_kw["ax_title"])

        # Axis
        caller(self.ax_axis, set_props_kw["axis"])

        # Axis Spines
        caller(self.ax_spines, set_props_kw["spines"])

        # Axis Labels
        caller(self.ax_labels, set_props_kw["labels"])

        # View (in case of 3D)
        caller(self.ax_view, set_props_kw["view"])

        # Grid
        caller(self.ax_grid, set_props_kw["grid"])

        # Legend
        self.ax_legend(set_props_kw["legend"], set_props_kw["legend_loc"])

        # Colorbar
        self.ax_colorbar(set_props_kw["colorbar"], **colorbar_kw)

        # Axis Limits
        caller(self.ax_limits, set_props_kw["limits"])

        # Axis Ticks
        caller(self.ax_ticks, set_props_kw["ticks"], set_props_kw["tick_labels"])

        # x-y Lines (through the origin)
        if set_props_kw["xy_lines"]:
            self.draw_xy_lines(**xy_lines_kw)

        # Face Color
        caller(self.ax_face_color, set_props_kw["face_color"])

        # -------------------------------------------------------

        # Save Figure
        file_name = None
        if save_file_name is not False:  # if user wants to save figure
            if save_file_name is True:  # default file name, otherwise provide a string
                save_file_name = None

            file_name = self.save_fig(file_name=save_file_name, **save_fig_kw)

        # Show Figure
        if set_props_kw["show_fig"]:
            self.show_fig()
            if set_props_kw["open_dir"]:
                open_file(os.path.split(file_name)[0])

        # Close Figure
        if set_props_kw["close_fig"]:
            plt.close(self.fig)

        return file_name


def new_figure(nrows=1, ncols=1,
        sharex=False, sharey=False,
        projection=None,
        squeeze=True,
        subplot_kw=None, gridspec_kw=None, **figure_kw) -> (Figure, Axes_plt):
    if subplot_kw is None:
        subplot_kw = dict()
    subplot_kw = {'projection': projection} | subplot_kw

    fig, axs = plt.subplots(nrows=nrows, ncols=ncols,
            sharex=sharex, sharey=sharey,
            squeeze=squeeze,
            subplot_kw=subplot_kw, gridspec_kw=gridspec_kw, **figure_kw)

    return fig, axs


# @copy_docstring_and_deprecators(_AxesLiron.draw_xy_lines)
def draw_xy_lines(ax: Axes_plt, **axis_lines_kw):
    _Axes(axs=ax).draw_xy_lines(**axis_lines_kw)


# @copy_docstring_and_deprecators(_AxesLiron.save_fig)
def save_fig(fig: Figure = None, file_name: str = None, **savefig_kw):
    if fig is None:
        fig = plt.gcf()

    _Axes(fig=fig).save_fig(file_name, **savefig_kw)


# @copy_docstring_and_deprecators(_AxesLiron.set_props)
def set_props(ax: Axes_plt = None,
        save_file_name: str | bool = False, save_fig_kw: dict = None,
        **set_props_kw):
    if ax is None:
        ax = plt.gca()

    _Axes(axs=ax).set_props(save_file_name=save_file_name,
            save_fig_kw=save_fig_kw,
            **set_props_kw)


def get_savefig_file_name(file_name: str = None, time_dir: bool = False, mkdir: bool = False):
    """

    Parameters
    ----------
    file_name :     str, optional
                    If <file_name> doesn't contain a directory name, the default saving directory
                    will be "<CUR_DIR>/figs", where <CUR_DIR> is the current directory of the directory
                    of the Python file being called.
    time_dir :      bool, optional
                    If True, a time directory will be created in the default saving directory.

    Returns
    -------
    Full path of file to be saved.
    """

    if file_name is None or os.path.dirname(file_name) == "":
        dir_name = os.path.join(MAIN_FILE_DIR, "figs")
    else:
        dir_name = os.path.dirname(file_name)

    if time_dir:
        dir_name = os.path.join(dir_name, TIME_STR)

    if mkdir and not os.path.exists(dir_name):
        mkdirs(dir_name)

    if file_name is None:
        file_name = os.path.join(dir_name, f"fig {get_time_str()}")
    elif os.path.dirname(file_name) == "":
        file_name = os.path.join(dir_name, file_name)

    return file_name



def hex2rgb(h):
	"""
	Convert hex color to RGB.

	Examples
	--------
	>>> hex2rgb("#FF5733")  # returns (255, 87, 51)

	Parameters
	----------
	h :             str or list of str
		Hex color string or list of hex color strings.

	Returns
	-------
	3-tuple or list of 3-tuples
	"""

	def hex2rgb_inner(h):
		return tuple(int(h.lstrip("#")[i:i + 2], 16) / 255 for i in (0, 2, 4))

	if isinstance(h, str):
		return hex2rgb_inner(h)

	return [hex2rgb_inner(hh) for hh in h]


def rgb2hex(rgb):
	"""
	Convert RGB color to hex.
	Examples
	--------
	>>> rgb2hex((255, 87, 51))  # returns "#FF5733"

	Parameters
	----------
	rgb :           3-tuple or list of 3-tuples

	Returns
	-------
	str or list of str
	"""

	def rgb2hex_inner(rgb):
		rgb = [int(c, 16) for c in rgb * 255]
		return "#{:02x}{:02x}{:02x}".format(*rgb).upper()

	if len(rgb) == 3 and isinstance(rgb[0], int):
		return rgb2hex_inner(rgb)

	return [rgb2hex_inner(r) for r in rgb]



from manim import *
import numpy as np


class BraceAnnotation(Scene):
    def construct(self):
        dot = Dot([-2, -1, 0])
        dot2 = Dot([2, 1, 0])
        line = Line(dot.get_center(), dot2.get_center()).set_color(ORANGE)
        b1 = Brace(line)
        b1text = b1.get_text("Horizontal distance")
        b2 = Brace(line, direction=line.copy().rotate(PI / 2).get_unit_vector())
        b2text = b2.get_tex("x-x_1")
        self.add(line, dot, dot2, b1, b2, b1text, b2text)


class OpeningManim(Scene):
    def construct(self):
        title = Tex(r"This is some \LaTeX")
        basel = MathTex(r"\sum_{n=1}^\infty \frac{1}{n^2} = \frac{\pi^2}{6}")
        VGroup(title, basel).arrange(DOWN)
        self.play(
            Write(title),
            FadeIn(basel, shift=DOWN),
        )
        self.wait()

        transform_title = Tex("That was a transform")
        transform_title.to_corner(UP + LEFT)
        self.play(
            Transform(title, transform_title),
            LaggedStart(*[FadeOut(obj, shift=DOWN) for obj in basel]),
        )
        self.wait()

        grid = NumberPlane()
        grid_title = Tex("This is a grid")
        grid_title.scale(1.5)
        grid_title.move_to(transform_title)

        self.add(grid, grid_title)  # Make sure title is on top of grid
        self.play(
            FadeOut(title),
            FadeIn(grid_title, shift=UP),
            Create(grid, run_time=3, lag_ratio=0.1),
        )
        self.wait()

        grid_transform_title = Tex(
            r"That was a non-linear function \\ applied to the grid"
        )
        grid_transform_title.move_to(grid_title, UL)
        grid.prepare_for_nonlinear_transform()
        self.play(
            grid.animate.apply_function(
                lambda p: p
                          + np.array(
                    [
                        np.sin(p[1]),
                        np.sin(p[0]),
                        0,
                    ]
                )
            ),
            run_time=3,
        )
        self.wait()
        self.play(Transform(grid_title, grid_transform_title))
        self.wait()

        self.renderer.file_writer.finish()


class MyTest(Scene):
    def construct(self):
        ax = Axes(x_range=[-5, 5, 1], y_range=[-5, 5, 1],
                  axis_config={'include_ticks': True, 'numbers_to_exclude': [0]}).add_coordinates()
        ax_labels = ax.get_axis_labels("x", "y")
        graph = ax.plot(lambda x: np.sin(x), x_range=[-1, 4])
        graph_label = ax.get_graph_label(graph, x_val=4)

        ax_staff = VGroup(ax, ax_labels)
        graph_staff = VGroup(graph, graph_label)

        self.play(Create(ax_staff, run_time=3))
        self.wait(3)
        self.play(Create(graph_staff, run_time=3))
        self.wait()

        box = Rectangle(stroke_color=MAROON, fill_color=BLUE_C, fill_opacity=0.5)

        self.play(
            Create(box, run_time=3)
        )
        self.play(box.animate.shift(3 * UP))


class AddToVGroup(Scene):
    def construct(self):
        circle_red = Circle(color=RED)
        circle_green = Circle(color=GREEN)
        circle_blue = Circle(color=BLUE)
        circle_red.shift(LEFT)
        circle_blue.shift(RIGHT)
        gr = VGroup(circle_red, circle_green)
        gr2 = VGroup(circle_blue)  # Constructor uses add directly
        self.add(gr, gr2)
        self.wait()
        gr += gr2  # Add group to another
        self.play(
            gr.animate.shift(DOWN),
        )
        gr -= gr2  # Remove group
        self.play(  # Animate groups separately
            gr.animate.shift(LEFT),
            gr2.animate.shift(UP),
        )
        self.play(  # Animate groups without modification
            (gr + gr2).animate.shift(RIGHT)
        )
        self.play(  # Animate group without component
            (gr - circle_red).animate.shift(RIGHT)
        )


if __name__ == "__main__":
    h = OpeningManim()
    h.construct()
    h.renderer.file_writer.finish()
    pass
